from typing import Union
import pulumi
import pulumi_databricks as databricks
from laktory.resourcesengines.pulumi.base import PulumiResourcesEngine
from laktory.models.group import Group
from laktory.models.user import User

from laktory._logger import get_logger

logger = get_logger(__name__)


class PulumiUser(PulumiResourcesEngine):
    @property
    def provider(self):
        return "databricks"

    def __init__(
        self,
        name=None,
        user: User = None,
        group_ids: dict[str, str] = None,
        opts=None,
    ):
        if name is None:
            name = f"user-{user.user_name}"
        super().__init__(self.t, name, {}, opts)

        opts = pulumi.ResourceOptions(
            parent=self,
            delete_before_replace=True,
        )

        if user.id is None:
            self.user = databricks.User(
                f"user-{user.user_name}",
                opts=opts,
                **user.model_pulumi_dump(),
            )
            user.id = self.user.id
        else:
            self.user = databricks.User.get(
                f"user-{user.user_name}",
                id=user.id,
                opts=opts,
                **user.model_pulumi_dump(),
            )

        self.roles = []
        for role in user.roles:
            self.roles += [
                databricks.UserRole(
                    f"user-role-{user.user_name}-{role}",
                    user_id=self.user.id,
                    role=role,
                    opts=opts,
                )
            ]

        if not group_ids:
            if user.groups:
                logger.warning(
                    "User is member of groups, but groups have not been provided. Group member resources will "
                    "be skipped."
                )
            return

        # Group Member
        self.group_members = []
        for g in user.groups:
            # Find matching group
            group_id = group_ids.get(g, None)

            if group_id:
                self.group_members += [
                    databricks.GroupMember(
                        f"group-member-{user.user_name}-{g}",
                        group_id=group_id,
                        member_id=self.user.id,
                        opts=opts,
                    )
                ]



from typing import Union

"""Provide several sample math calculations.

This module allows the user to make mathematical calculations.

Examples:
    >>> from calculator import calculations
    >>> calculations.add(2, 4)
    6.0
    >>> calculations.multiply(2.0, 4.0)
    8.0
    >>> from calculator.calculations import divide
    >>> divide(4.0, 2)
    2.0

The module contains the following functions:

- `add(a, b)` - Returns the sum of two numbers.
- `subtract(a, b)` - Returns the difference of two numbers.
- `multiply(a, b)` - Returns the product of two numbers.
- `divide(a, b)` - Returns the quotient of two numbers.
"""


def add(a: Union[float, int], b: Union[float, int]) -> float:
    """
    Compute and return the sum of two numbers.

    Examples
    --------
    >>> add(4.0, 2.0)
    6.0
    >>> add(4, 2)
    6.0

    Parameters
    ----------
    a: A number representing the first addend in the addition.
    b: A number representing the second addend in the addition.

    Returns
    -------
    A number representing the arithmetic sum of `a` and `b`.
    """
    return float(a + b)


def subtract(a: Union[float, int], b: Union[float, int]) -> float:
    """Calculate the difference of two numbers.

    Examples:
        >>> subtract(4.0, 2.0)
        2.0
        >>> subtract(4, 2)
        2.0

    Args:
        a: A number representing the minuend in the subtraction.
        b: A number representing the subtrahend in the subtraction.

    Returns:
        A number representing the difference between `a` and `b`.
    """
    return float(a - b)


def multiply(a: Union[float, int], b: Union[float, int]) -> float:
    """Compute and return the product of two numbers.

    Examples:
        >>> multiply(4.0, 2.0)
        8.0
        >>> multiply(4, 2)
        8.0

    Args:
        a: A number representing the multiplicand in the multiplication.
        b: A number representing the multiplier in the multiplication.

    Returns:
        A number representing the product of `a` and `b`.
    """
    return float(a * b)


def divide(a: Union[float, int], b: Union[float, int]) -> float:
    """Compute and return the quotient of two numbers.

    Examples:
        >>> divide(4.0, 2.0)
        2.0
        >>> divide(4, 2)
        2.0
        >>> divide(4, 0)
        Traceback (most recent call last):
        ...
        ZeroDivisionError: division by zero

    Args:
        a: A number representing the dividend in the division.
        b: A number representing the divisor in the division.

    Returns:
        A number representing the quotient of `a` and `b`.

    Raises:
        ZeroDivisionError: An error occurs when the divisor is `0`.
    """
    if b == 0:
        raise ZeroDivisionError("division by zero")
    return float(a / b)



from laktory.models import Table
from laktory.models import Schema

schema = Schema(
    name="stocks",
    catalog_name="default",
    tables=[
        Table(
            name="AAPL",
            columns=[
                {
                    "name": "open",
                    "type": "double",
                },
                {
                    "name": "close",
                    "type": "double",
                },
            ],
        ),
        Table(
            name="GOOGL",
            columns=[
                {
                    "name": "${var.dynamic_column}",
                    "type": "double",
                },
                {
                    "name": "high",
                    "type": "double",
                },
            ],
        ),
    ],
    _vars={"dynamic_column": "low"},
)


def test_inject_vars():
    d0 = schema.model_dump()
    d1 = schema.inject_vars(d0)
    assert d1["tables"][-1]["columns"][0]["name"] == "low"


def test_pulumi_dump():
    d = schema.model_pulumi_dump()
    assert d == {
        "name": "stocks",
        "comment": None,
        "catalog_name": "default",
        "force_destroy": True,
    }


if __name__ == "__main__":
    test_inject_vars()
    test_pulumi_dump()



from typing import Union

from laktory.models.base import BaseModel
from laktory.models.resources import Resources
from laktory.models.sql.schema import Schema
from laktory.models.grants.cataloggrant import CatalogGrant


class Catalog(BaseModel, Resources):
    name: str
    comment: Union[str, None] = None
    schemas: list[Schema] = []
    is_unity: bool = True
    grants: list[CatalogGrant] = None

    # Deployment options
    owner: str = None
    force_destroy: bool = True
    storage_root: str = None
    isolation_mode: str = "OPEN"

    def model_post_init(self, __context):
        for schema in self.schemas:
            schema.catalog_name = self.name
            schema.model_post_init(None)

    # ----------------------------------------------------------------------- #
    # Computed fields                                                         #
    # ----------------------------------------------------------------------- #

    @property
    def full_name(self) -> str:
        return self.name

    # ----------------------------------------------------------------------- #
    # Methods                                                                 #
    # ----------------------------------------------------------------------- #
    #
    # TODO: Move to Databricks SDK engine
    # def exists(self):
    #     return self.name in [c.name for c in self.workspace_client.catalogs.list()]
    #
    # def create(self, if_not_exists: bool = True):
    #     w = self.workspace_client
    #     exists = self.exists()
    #
    #     if if_not_exists and exists:
    #         return w.catalogs.get(self.name)
    #
    #     return w.catalogs.create(
    #         name=self.name,
    #         comment=self.comment,
    #     )
    #
    # def delete(self, force: bool = False):
    #     self.workspace_client.catalogs.delete(self.name, force=force)

    # ----------------------------------------------------------------------- #
    # Resources Engine Methods                                                #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_excludes(self) -> list[str]:
        return ["schemas", "is_unity", "grants"]

    def deploy_with_pulumi(self, name=None, opts=None):
        from laktory.resourcesengines.pulumi.catalog import PulumiCatalog

        return PulumiCatalog(name=name, catalog=self, opts=opts)



from datetime import datetime

from laktory.models import Catalog


def test_model():
    cat = Catalog(
        name="lakehouse",
    )

    assert cat.name == "lakehouse"
    assert cat.full_name == "lakehouse"


def atest_create():
    # TODO: Remove - deprecated
    # Timestamp is included in catalog name to prevent conflicts when running
    # multiple tests in parallel
    catalog_name = "laktory_testing_" + str(datetime.now().timestamp()).replace(".", "")

    cat = Catalog(name=catalog_name)
    cat.create(if_not_exists=True)
    assert cat.exists()
    cat.delete(force=True)
    assert not cat.exists()


if __name__ == "__main__":
    test_model()
    atest_create()



from typing import Optional
from typing import Union
from pydantic import Field
from pydantic import field_validator
from pydantic_core.core_schema import FieldValidationInfo

from laktory._settings import settings
from laktory.models.base import BaseModel
from laktory.models.producer import Producer


class DataEventHeader(BaseModel):
    name: str = Field(...)
    description: Union[str, None] = Field(None)
    producer: Producer = Field(None)
    events_root: str = settings.workspace_landing_root + "events/"

    # ----------------------------------------------------------------------- #
    # Paths                                                                   #
    # ----------------------------------------------------------------------- #

    @property
    def event_root(self):
        producer = ""
        if self.producer is not None:
            producer = self.producer.name + "/"
        v = f"{self.events_root}{producer}{self.name}/"
        return v

    #
    # @field_validator("root_path")
    # def default_root_path(cls, v: str, info: FieldValidationInfo) -> str:
    #     if v is None:
    #         data = info.data
    #         producer = ""
    #         if data.get("producer") is not None:
    #             producer = data["producer"].name + "/"
    #         v = f'{data.get("events_root_path", "")}{producer}{data.get("name", "")}/'
    #     return v



from typing import Union
from laktory.models.base import BaseModel
from laktory.models.resources import Resources


class Group(BaseModel, Resources):
    display_name: str
    allow_cluster_create: bool = False
    workspace_access: bool = None
    user_names: list[str] = None
    id: Union[str, None] = None

    # ----------------------------------------------------------------------- #
    # Resources Engine Methods                                                #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_excludes(self) -> list[str]:
        return ["user_names", "id"]

    def deploy_with_pulumi(self, name=None, opts=None):
        from laktory.resourcesengines.pulumi.group import PulumiGroup

        return PulumiGroup(name=name, group=self, opts=opts)



from laktory.models import User
from laktory.models import Group

user = User(
    user_name="data.engineer@gmail.moc",
    groups=[
        "role-engineers",
    ],
    roles=["store-admin"],
)

group = Group(
    display_name="role-engineers",
    user_names=[
        "data.engineer@gmail.moc",
    ],
)


def test_user_group():
    assert user.roles == ["store-admin"]
    assert group.user_names == ["data.engineer@gmail.moc"]


if __name__ == "__main__":
    test_user_group()



from typing import Literal
from laktory.models.base import BaseModel
from laktory.models.grants.schemagrant import PRIVILEGES as DB_PRIVILEGES

PRIVILEGES = DB_PRIVILEGES + [
    "USE_CATALOG",
    "CREATE_SCHEMA",
]


class CatalogGrant(BaseModel):
    principal: str
    privileges: list[Literal[tuple(PRIVILEGES)]]



from typing import Literal
from laktory.models.base import BaseModel

PRIVILEGES = [
    "SELECT",
]


class ShareGrant(BaseModel):
    principal: str
    privileges: list[Literal[tuple(PRIVILEGES)]]



import pulumi
import pulumi_databricks as databricks
from laktory.resourcesengines.pulumi.base import PulumiResourcesEngine
from laktory.models.group import Group


class PulumiGroup(PulumiResourcesEngine):
    @property
    def provider(self):
        return "databricks"

    def __init__(
        self,
        name=None,
        group: Group = None,
        opts=None,
    ):
        if name is None:
            name = f"group-{group.display_name}"
        super().__init__(self.t, name, {}, opts)

        opts = pulumi.ResourceOptions(
            parent=self,
            delete_before_replace=True,
        )

        if group.id is None:
            self.group = databricks.Group(
                f"group-{group.display_name}", opts=opts, **group.model_pulumi_dump()
            )
            group.id = self.group.id
        else:
            self.group = databricks.Group.get(
                f"group-{group.display_name}", id=group.id
            )



from typing import Union

from laktory._logger import get_logger
from laktory.models.base import BaseModel

logger = get_logger(__name__)


class OrderBy(BaseModel):
    sql_expression: str
    desc: bool = False


class TableWindowFilter(BaseModel):
    partition_by: Union[list[str], None]
    order_by: Union[list[OrderBy], None] = None
    descending: bool = True
    rows_to_keep: int = 1
    drop_row_index: bool = True
    row_index_name: str = "_row_index"

    def run(self, df):
        from pyspark.sql import Window
        import pyspark.sql.functions as F

        # Partition by
        w = Window.partitionBy(*self.partition_by)

        # Order by
        if self.order_by:
            order_bys = []
            for o in self.order_by:
                order_bys += [F.expr(o.sql_expression)]
                if o.desc:
                    order_bys[-1] = order_bys[-1].desc()
            w = w.orderBy(*order_bys)

        # Set rows index
        df = df.withColumn(self.row_index_name, F.row_number().over(w))

        # Filter
        df = df.filter(F.col(self.row_index_name) <= self.rows_to_keep)
        if self.drop_row_index:
            df = df.drop(self.row_index_name)

        return df



from typing import Union

from laktory.models.base import BaseModel
from laktory.models.resources import Resources
from laktory.models.sql.table import Table
from laktory.models.sql.volume import Volume
from laktory.models.grants.schemagrant import SchemaGrant


class Schema(BaseModel, Resources):
    name: str
    comment: Union[str, None] = None
    tables: list[Table] = []
    volumes: list[Volume] = []
    catalog_name: Union[str, None] = None
    grants: list[SchemaGrant] = None

    # Deployment options
    force_destroy: bool = True

    def model_post_init(self, __context):
        for table in self.tables:
            table.catalog_name = self.catalog_name
            table.schema_name = self.name
        for volume in self.volumes:
            volume.catalog_name = self.catalog_name
            volume.schema_name = self.name

    # ----------------------------------------------------------------------- #
    # Computed fields                                                         #
    # ----------------------------------------------------------------------- #

    @property
    def parent_full_name(self) -> str:
        return self.catalog_name

    @property
    def full_name(self) -> str:
        _id = self.name
        if self.parent_full_name is not None:
            _id = f"{self.parent_full_name}.{_id}"
        return _id

    # ----------------------------------------------------------------------- #
    # Methods                                                                 #
    # ----------------------------------------------------------------------- #
    # TODO: Move to Databricks SDK engine
    # def exists(self):
    #     return self.name in [
    #         c.name
    #         for c in self.workspace_client.schemas.list(catalog_name=self.catalog_name)
    #     ]
    #
    # def create(self, if_not_exists: bool = True):
    #     w = self.workspace_client
    #     exists = self.exists()
    #
    #     if if_not_exists and exists:
    #         return w.schemas.get(self.full_name)
    #
    #     return w.schemas.create(
    #         name=self.name,
    #         catalog_name=self.catalog_name,
    #         comment=self.comment,
    #     )
    #
    # def delete(self):
    #     self.workspace_client.schemas.delete(self.full_name)

    # ----------------------------------------------------------------------- #
    # Resources Engine Methods                                                #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_excludes(self) -> list[str]:
        return ["tables", "volumes", "grants"]

    def deploy_with_pulumi(self, name=None, opts=None):
        from laktory.resourcesengines.pulumi.schema import PulumiSchema

        return PulumiSchema(name=name, schema=self, opts=opts)



from typing import Union

import pulumi
import pulumi_databricks as databricks
from laktory.resourcesengines.pulumi.base import PulumiResourcesEngine
from laktory.models.sql.schema import Schema

from laktory._logger import get_logger

logger = get_logger(__name__)


class PulumiSchema(PulumiResourcesEngine):
    @property
    def provider(self):
        return "databricks"

    def __init__(
        self,
        name=None,
        schema: Schema = None,
        opts=None,
    ):
        if name is None:
            name = f"schema-{schema.full_name}"
        super().__init__(self.t, name, {}, opts)

        opts = pulumi.ResourceOptions(
            parent=self,
        )

        # Schema
        self.schema = databricks.Schema(
            f"schema-{schema.full_name}",
            opts=opts,
            **schema.model_pulumi_dump(),
        )

        # Schema grants
        _opts = opts.merge(pulumi.ResourceOptions(depends_on=self.schema))
        if schema.grants:
            self.grants = databricks.Grants(
                f"grants-{schema.full_name}",
                schema=schema.full_name,
                grants=[
                    databricks.GrantsGrantArgs(
                        principal=g.principal, privileges=g.privileges
                    )
                    for g in schema.grants
                ],
                opts=_opts,
            )

        # Schema volumes
        if schema.volumes:
            for v in schema.volumes:
                v.vars = schema.vars
                v._resources = v.deploy_with_pulumi(
                    opts=pulumi.ResourceOptions(parent=self.schema)
                )

        # TODO: Schema tables



import os
from typing import Any
from typing import Literal
from pydantic import model_validator
from laktory.models.base import BaseModel
from laktory.models.resources import Resources
from laktory.models.permission import Permission


class Notebook(BaseModel, Resources):
    source: str
    dirpath: str = None
    path: str = None
    language: Literal["SCALA", "PYTHON", "SQL", "R"] = None
    permissions: list[Permission] = []

    @property
    def filename(self):
        return os.path.basename(self.source)

    @property
    def key(self):
        key = os.path.splitext(self.path)[0].replace("/", "-")
        if key.startswith("-"):
            key = key[1:]
        return key

    @model_validator(mode="after")
    def default_path(self) -> Any:
        if self.path is None:
            if self.dirpath:
                self.path = f"{self.dirpath}{self.filename}"

            elif "/notebooks/" in self.source:
                self.path = "/" + self.source.split("/notebooks/")[-1]

        return self

    # ----------------------------------------------------------------------- #
    # Resources Engine Methods                                                #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_excludes(self) -> list[str]:
        return ["permissions", "dirpath"]

    def deploy_with_pulumi(self, name=None, groups=None, opts=None):
        from laktory.resourcesengines.pulumi.notebook import PulumiNotebook

        return PulumiNotebook(name=name, notebook=self, opts=opts)



# # --------------------------------------------------------------------------- #
# # compare                                                                     #
# # --------------------------------------------------------------------------- #
#
# @pandas_udf(sqlt.BooleanType(), udfuncs)
# def compare(
#         x: pd.Series,
#         y: Union[float, pd.Series] = 0,
#         where: pd.Series = None,
#         operator: str = "==",
#         default: bool = None,
# ) -> pd.Series:
#     """
#     Compare a series `x` and a value or another series `y` using
#     `operator`. Comparison can be limited to `where` and assigned
#     `default` elsewhere.
#
#     output = `x` `operator` `y`
#
#     Parameters
#     ---------
#     x : pd.Series
#         Base series to compare
#     y : pd.Series or Object
#         Series or object to compare to
#     where: pd.Series
#         Where to apply the comparison
#     operator: str
#         Operator for comparison
#     default: bool
#         Default value to be applied outside of where
#
#     Returns
#     -------
#     output: pd.Series
#         Comparison result
#     """
#     output = pd.Series(default, x.index)
#
#     if where is None:
#         where = pd.Series(True, x.index)
#     elif isinstance(where, pd.Series):
#         where = where.astype(bool)
#
#     if not isinstance(y, pd.Series):
#         y = pd.Series(y, index=x.index)
#
#     if operator == "==":
#         output.loc[where] = x.loc[where] == y.loc[where]
#     elif operator == "!=":
#         output.loc[where] = x.loc[where] != y.loc[where]
#     elif operator == "<":
#         output.loc[where] = x.loc[where] < y.loc[where]
#     elif operator == "<=":
#         output.loc[where] = x.loc[where] <= y.loc[where]
#     elif operator == ">":
#         output.loc[where] = x.loc[where] > y.loc[where]
#     elif operator == ">=":
#         output.loc[where] = x.loc[where] >= y.loc[where]
#     else:
#         raise ValueError(f"Operator '{operator}' is not supported.")
#
#     return output
#



from typing import Union

from laktory.models.base import BaseModel


class Producer(BaseModel):
    name: str
    description: Union[str, None] = None
    party: int = 1  # First, Second and Third party data
    # ref.: https://blog.hubspot.com/service/first-party-data



from typing import Union

import pulumi
import pulumi_databricks as databricks
from laktory.resourcesengines.pulumi.base import PulumiResourcesEngine
from laktory.models.sql.catalog import Catalog

from laktory._logger import get_logger

logger = get_logger(__name__)


class PulumiCatalog(PulumiResourcesEngine):
    @property
    def provider(self):
        return "databricks"

    def __init__(
        self,
        name=None,
        catalog: Catalog = None,
        opts=None,
    ):
        if name is None:
            name = f"catalog-{catalog.full_name}"
        super().__init__(self.t, name, {}, opts)

        opts = pulumi.ResourceOptions(parent=self)

        # Catalog
        self.catalog = databricks.Catalog(
            f"catalog-{catalog.full_name}",
            opts=opts,
            **catalog.model_pulumi_dump(),
        )

        # Grants
        if catalog.grants:
            self.grants = databricks.Grants(
                f"grants-catalog-{catalog.full_name}",
                catalog=self.catalog.name,
                grants=[
                    databricks.GrantsGrantArgs(
                        principal=g.principal, privileges=g.privileges
                    )
                    for g in catalog.grants
                ],
                opts=opts,
            )

        # Schemas
        if catalog.schemas:
            for s in catalog.schemas:
                s.vars = catalog.vars
                s._resources = s.deploy_with_pulumi(
                    opts=pulumi.ResourceOptions(parent=self.catalog)
                )



from typing import Literal
from laktory.models.base import BaseModel

PRIVILEGES = [
    "ALL_PRIVILEGES",
    "CREATE_FOREIGN_CATALOG",
    "USE_CONNECTION",
]


class ConnectionGrant(BaseModel):
    principal: str
    privileges: list[Literal[tuple(PRIVILEGES)]]



# from laktory.sql import py_to_sql
# TODO: Delete (deprecated)


def atest_sql():
    # String
    s = py_to_sql("test")
    assert s == "'test'"

    # Number
    s = py_to_sql(3)
    assert s == "3"

    # Boolean
    s = py_to_sql(False)
    assert s == "False"

    # Dict
    s = py_to_sql(
        {
            "name": "John",
            "surname": "Doe",
            "address": {"number": 0, "street": "Main"},
            "verified": True,
        }
    )
    assert (
        s
        == "named_struct('name', 'John', 'surname', 'Doe', 'address', named_struct('number', 0, 'street', 'Main'), 'verified', True)"
    )

    # List
    s = py_to_sql([1, "test", True])
    assert s == "ARRAY(1, 'test', True)"

    # List of dicts
    s = py_to_sql(
        [
            {"numbers": [1, 2, 3], "address": {"number": 0, "street": "Main"}},
            True,
        ]
    )
    assert (
        s
        == "ARRAY(named_struct('numbers', ARRAY(1, 2, 3), 'address', named_struct('number', 0, 'street', 'Main')), True)"
    )


if __name__ == "__main__":
    atest_sql()



from typing import Any
from typing import Union

from pydantic import model_validator

from laktory._logger import get_logger
from laktory.models.base import BaseModel
from laktory.models.sql.column import Column
from laktory.models.compute.tablebuilder import TableBuilder
from laktory.models.grants.tablegrant import TableGrant

logger = get_logger(__name__)


class Table(BaseModel):
    catalog_name: Union[str, None] = None
    columns: list[Column] = []
    comment: Union[str, None] = None
    data: list[list[Any]] = None
    grants: list[TableGrant] = None
    name: str
    primary_key: Union[str, None] = None
    schema_name: Union[str, None] = None
    timestamp_key: Union[str, None] = None
    builder: TableBuilder = TableBuilder()

    # ----------------------------------------------------------------------- #
    # Validators                                                              #
    # ----------------------------------------------------------------------- #

    @model_validator(mode="after")
    def assign_catalog_schema(self) -> Any:
        # Assign to columns
        for c in self.columns:
            c.table_name = self.name
            c.catalog_name = self.catalog_name
            c.schema_name = self.schema_name

        # Set builder table
        self.builder._table = self

        # Assign to sources
        if self.builder.table_source is not None:
            if self.builder.table_source.catalog_name is None:
                self.builder.table_source.catalog_name = self.catalog_name
            if self.builder.table_source.schema_name is None:
                self.builder.table_source.schema_name = self.schema_name

        # Assign to joins
        for join in self.builder.joins:
            if join.other.catalog_name is None:
                join.other.catalog_name = self.catalog_name
            if join.other.schema_name is None:
                join.other.schema_name = self.schema_name

        return self

    # ----------------------------------------------------------------------- #
    # Computed fields                                                         #
    # ----------------------------------------------------------------------- #

    @property
    def parent_full_name(self) -> str:
        _id = ""
        if self.catalog_name:
            _id += self.catalog_name

        if self.schema_name:
            if _id == "":
                _id = self.schema_name
            else:
                _id += f".{self.schema_name}"

        return _id

    @property
    def full_name(self) -> str:
        _id = self.name
        if self.parent_full_name is not None:
            _id = f"{self.parent_full_name}.{_id}"
        return _id

    @property
    def database_name(self) -> str:
        return self.schema_name

    @property
    def zone(self):
        return self.builder.zone

    # ----------------------------------------------------------------------- #
    # Properties                                                              #
    # ----------------------------------------------------------------------- #

    @property
    def column_names(self):
        return [c.name for c in self.columns]

    @property
    def df(self):
        import pandas as pd

        return pd.DataFrame(data=self.data, columns=self.column_names)

    def to_df(self, spark=None):
        import pandas as pd

        df = pd.DataFrame(data=self.data, columns=self.column_names)

        if spark:
            df = spark.createDataFrame(df)
        return df

    @property
    def is_from_cdc(self):
        return self.builder.is_from_cdc



import os
from typing import Any
from typing import Literal
from typing import Union
from pydantic import model_validator
from pydantic import Field
from laktory.models.base import BaseModel
from laktory.models.resources import Resources
from laktory.models.permission import Permission
from laktory.models.compute.cluster import Cluster
from laktory.models.sql.table import Table


class PipelineLibraryFile(BaseModel):
    path: str


class PipelineLibraryNotebook(BaseModel):
    path: str


class PipelineLibrary(BaseModel):
    file: str = None
    notebook: PipelineLibraryNotebook = None


class PipelineNotifications(BaseModel):
    alerts: list[
        Literal[
            "on-update-success",
            "on-update-failure",
            "on-update-fatal-failure",
            "on-flow-failure",
        ]
    ]
    recipients: list[str]


class PipelineCluster(Cluster):
    autotermination_minutes: int = Field(None)
    cluster_id: str = Field(None)
    data_security_mode: str = Field(None)
    enable_elastic_disk: bool = Field(None)
    idempotency_token: str = Field(None)
    is_pinned: bool = Field(None)
    libraries: list[Any] = Field(None)
    node_type_id: str = None
    runtime_engine: str = Field(None)
    single_user_name: str = Field(None)
    spark_version: str = Field(None)

    @model_validator(mode="after")
    def excluded_fields(self) -> Any:
        for f in [
            "autotermination_minutes",
            "cluster_id",
            "data_security_mode",
            "enable_elastic_disk",
            "idempotency_token",
            "is_pinned",
            "libraries",
            "runtime_engine",
            "single_user_name",
            "spark_version",
        ]:
            if getattr(self, f, None) not in [None, [], {}]:
                raise ValueError(f"Field {f} should be null")

        return self


class PipelineUDF(BaseModel):
    module_name: str
    function_name: str
    module_path: str = None


class Pipeline(BaseModel, Resources):
    allow_duplicate_names: bool = None
    catalog: str = None
    channel: str = "PREVIEW"
    clusters: list[PipelineCluster] = []
    configuration: dict[str, str] = {}
    continuous: bool = None
    development: Union[bool, str] = None
    edition: str = None
    # filters
    libraries: list[PipelineLibrary] = []
    name: str
    notifications: list[PipelineNotifications] = []
    permissions: list[Permission] = []
    photon: bool = None
    serverless: bool = None
    storage: str = None
    tables: list[Table] = []
    target: str = None
    udfs: list[PipelineUDF] = []

    @model_validator(mode="after")
    def assign_pipeline_to_tables(self) -> Any:
        for t in self.tables:
            t.builder.pipeline_name = self.name
            t.catalog_name = self.catalog
            t.schema_name = self.target
            for c in t.columns:
                c.table_name = t.name
                c.catalog_name = t.catalog_name
                c.schema_name = t.schema_name

            # Assign to sources
            if t.builder.table_source is not None:
                if t.builder.table_source.catalog_name is None:
                    t.builder.table_source.catalog_name = self.catalog
                if t.builder.table_source.schema_name is None:
                    t.builder.table_source.schema_name = self.target

        return self

    # ----------------------------------------------------------------------- #
    # Resources Engine Methods                                                #
    # ----------------------------------------------------------------------- #

    @property
    def id(self):
        if self._resources is None:
            return None
        return self.resources.pipeline.id

    @property
    def pulumi_excludes(self) -> list[str]:
        return {
            "permissions": True,
            "tables": True,
            "clusters": {"__all__": {"permissions"}},
            "udfs": True,
        }

    def model_pulumi_dump(self, *args, **kwargs):
        d = super().model_pulumi_dump(*args, **kwargs)
        _clusters = []
        for c in d.get("clusters", []):
            c["label"] = c.pop("name")
            _clusters += [c]
        d["clusters"] = _clusters
        return d

    def deploy_with_pulumi(self, name=None, groups=None, opts=None):
        from laktory.resourcesengines.pulumi.pipeline import PulumiPipeline

        return PulumiPipeline(name=name, pipeline=self, opts=opts)

    # ----------------------------------------------------------------------- #
    # Methods                                                                 #
    # ----------------------------------------------------------------------- #
    # TODO: Move to Databricks SDK engine
    # def get_tables_meta(self, catalog_name="main", schema_name="laktory") -> Table:
    #     table = Table.meta_table()
    #     table.catalog_name = catalog_name
    #     table.schema_name = schema_name
    #
    #     data = []
    #     for t in self.tables:
    #         _dump = t.model_dump(mode="json")
    #         _data = []
    #         for c in table.column_names:
    #             _data += [_dump[c]]
    #         data += [_data]
    #     table.data = data
    #
    #     return table
    #
    # def get_columns_meta(self, catalog_name="main", schema_name="laktory") -> Table:
    #     table = Column.meta_table()
    #     table.catalog_name = catalog_name
    #     table.schema_name = schema_name
    #
    #     data = []
    #     for t in self.tables:
    #         for c in t.columns:
    #             _dump = c.model_dump(mode="json")
    #             _data = []
    #             for k in table.column_names:
    #                 _data += [_dump[k]]
    #             data += [_data]
    #     table.data = data
    #
    #     return table
    #
    # def publish_tables_meta(self, catalog_name="main", schema_name="laktory", init=True):
    #
    #     # Create catalog
    #     Catalog(name=catalog_name).create(if_not_exists=True)
    #
    #     # Create schema
    #     Schema(name=schema_name, catalog_name=catalog_name).create(if_not_exists=True)
    #
    #     # Get and create tables
    #     tables = self.get_tables_meta(
    #         catalog_name=catalog_name, schema_name=schema_name
    #     )
    #     tables.create(or_replace=init, insert_data=True)
    #
    #     # Get and create tables
    #     columns = self.get_columns_meta(
    #         catalog_name=catalog_name, schema_name=schema_name
    #     )
    #     columns.create(or_replace=init, insert_data=True)



import re
import json
from pyspark.sql.dataframe import DataFrame


def schema_flat(df: DataFrame) -> list[str]:
    def get_fields(json_schema):
        field_names = []
        for f in json_schema.get("fields", []):
            f_name = f["name"]
            f_type = f["type"]
            if isinstance(f_type, dict):
                if f_type["type"] == "array":
                    e_type = f_type["elementType"]
                    field_names += [f_name]
                    if isinstance(e_type, dict):
                        _field_names = get_fields(f_type["elementType"])
                        field_names += [f"{f_name}[*].{v}" for v in _field_names]
                elif f_type["type"] == "struct":
                    _field_names = get_fields(f["type"])
                    field_names += [f_name]
                    field_names += [f"{f_name}.{v}" for v in _field_names]
                elif f_type["type"] == "map":
                    field_names += [f_name]
                else:
                    raise ValueError(f_type["type"])
            else:
                field_names += [f_name]
        return field_names

    return get_fields(json.loads(df.schema.json()))


def has_column(df: DataFrame, col: str) -> bool:
    _col = re.sub(r"\[(\d+)\]", r"[*]", col)
    _col = re.sub(r"`", "", _col)
    return _col in schema_flat(df)


# DataFrame Extensions
DataFrame.schema_flat = schema_flat
DataFrame.has_column = has_column

# Spark Connect DataFrame Extensions
try:
    from pyspark.sql.connect.dataframe import DataFrame

    DataFrame.schema_flat = schema_flat
    DataFrame.has_column = has_column
except:
    pass



from pydantic import model_validator
from typing import Any
from typing import Literal
from typing import Union

from laktory._logger import get_logger
from laktory.models.compute.tablejoin import TableJoin
from laktory.models.base import BaseModel
from laktory.models.sql.column import Column
from laktory.spark import DataFrame
from laktory.models.datasources import TableDataSource
from laktory.models.datasources import EventDataSource
from laktory.models.compute.tableaggregation import TableAggregation
from laktory.models.compute.tablewindowfilter import TableWindowFilter

logger = get_logger(__name__)


class TableBuilder(BaseModel):
    aggregation: Union[TableAggregation, None] = None
    drop_source_columns: Union[bool, None] = None
    drop_duplicates: Union[bool, None] = None
    drop_columns: list[str] = []
    event_source: Union[EventDataSource, None] = None
    filter: Union[str, None] = None
    joins: list[TableJoin] = []
    joins_post_aggregation: list[TableJoin] = []
    pipeline_name: Union[str, None] = None
    selects: Union[list[str], dict[str, str], None] = None
    table_source: Union[TableDataSource, None] = None
    template: Union[str, bool, None] = None
    window_filter: Union[TableWindowFilter, None] = None
    zone: Literal["BRONZE", "SILVER", "SILVER_STAR", "GOLD"] = None
    _table: Any = None
    _columns_to_build = []

    @model_validator(mode="after")
    def default_options(self) -> Any:
        # Default values
        if self.zone == "BRONZE":
            if self.drop_source_columns is None:
                self.drop_source_columns = False
            if self.drop_duplicates is not None:
                self.drop_duplicates = False

        if self.zone == "SILVER":
            if self.drop_source_columns is None:
                self.drop_source_columns = True
            if self.drop_duplicates is not None:
                self.drop_duplicates = True

        if self.zone == "SILVER_STAR":
            if self.drop_source_columns is None:
                self.drop_source_columns = False
            if self.drop_duplicates is not None:
                self.drop_duplicates = False

        if self.zone == "GOLD":
            if self.drop_source_columns is None:
                self.drop_source_columns = False
            if self.drop_duplicates is not None:
                self.drop_duplicates = False

        if self.template is None:
            self.template = self.zone

        return self

    @property
    def source(self):
        if self.event_source is not None and self.event_source.name is not None:
            return self.event_source
        elif self.table_source is not None and self.table_source.name is not None:
            return self.table_source

    @property
    def is_from_cdc(self):
        if self.source is None:
            return False
        else:
            return self.source.is_cdc

    @property
    def columns(self):
        return self._table.columns

    @property
    def timestamp_key(self):
        return self._table.timestamp_key

    @property
    def primary_key(self):
        return self._table.primary_key

    @property
    def has_joins(self):
        return len(self.joins) > 0

    @property
    def has_joins_post_aggregation(self):
        return len(self.joins_post_aggregation) > 0

    def get_zone_columns(self, zone, df=None):
        from laktory.spark.dataframe import has_column

        cols = []

        if zone == "BRONZE":
            cols = [
                Column(
                    **{
                        "name": "_bronze_at",
                        "type": "timestamp",
                        "spark_func_name": "current_timestamp",
                    }
                )
            ]

        elif zone == "SILVER":
            if self.timestamp_key:
                cols += [
                    Column(
                        **{
                            "name": "_tstamp",
                            "type": "timestamp",
                            "spark_func_name": "coalesce",
                            "spark_func_args": [self.timestamp_key],
                        }
                    )
                ]

            if has_column(df, "_bronze_at"):
                cols += [
                    Column(
                        **{
                            "name": "_bronze_at",
                            "type": "timestamp",
                            "spark_func_name": "coalesce",
                            "spark_func_args": ["_bronze_at"],
                        }
                    )
                ]

            cols += [
                Column(
                    **{
                        "name": "_silver_at",
                        "type": "timestamp",
                        "spark_func_name": "current_timestamp",
                    }
                )
            ]

        elif zone == "SILVER_STAR":
            cols = [
                Column(
                    **{
                        "name": "_silver_star_at",
                        "type": "timestamp",
                        "spark_func_name": "current_timestamp",
                    }
                )
            ]

        elif zone == "GOLD":
            cols = [
                Column(
                    **{
                        "name": "_gold_at",
                        "type": "timestamp",
                        "spark_func_name": "current_timestamp",
                    }
                )
            ]

        return cols

    def read_source(self, spark) -> DataFrame:
        return self.source.read(spark)

    def build_columns(self, df, udfs=None, raise_exception=True) -> DataFrame:
        logger.info(f"Setting columns...")
        built_cols = []
        for col in self._columns_to_build:
            c = col.to_spark(df, udfs=udfs, raise_exception=raise_exception)
            if c is not None:
                df = df.withColumn(col.name, c)
                built_cols += [col]

        for c in built_cols:
            self._columns_to_build.remove(c)

        return df

    def process(self, df, udfs=None, spark=None) -> DataFrame:
        import pyspark.sql.functions as F

        logger.info(f"Applying {self.zone} transformations")

        # Build columns
        self._columns_to_build = self.columns + self.get_zone_columns(
            zone=self.zone, df=df
        )
        column_names = [c.name for c in self._columns_to_build]
        df = self.build_columns(
            df, udfs=udfs, raise_exception=not (self.has_joins or self.aggregation)
        )

        # Make joins
        for i, join in enumerate(self.joins):
            if i == 0:
                name = self.source.name
            else:
                name = "previous_join"
            join.left = TableDataSource(name=name)
            join.left._df = df
            df = join.run(spark)

            # Build remaining columns again (in case inputs are found in joins)
            df = self.build_columns(
                df, udfs=udfs, raise_exception=i == len(self.joins) - 1
            )

        # Window filtering
        if self.window_filter:
            df = self.window_filter.run(df)

        # Drop source columns
        if self.drop_source_columns:
            logger.info(f"Dropping source columns...")
            df = df.select(column_names)

        if self.aggregation:
            df = self.aggregation.run(df, udfs=udfs)
            self._columns_to_build += self.get_zone_columns(zone=self.zone, df=df)

        # Build columns after aggregation
        df = self.build_columns(
            df, udfs=udfs, raise_exception=not self.has_joins_post_aggregation
        )

        # Make post-aggregation joins
        for i, join in enumerate(self.joins_post_aggregation):
            if i == 0:
                name = self.source.name
            else:
                name = "previous_join"
            join.left = TableDataSource(name=name)
            join.left._df = df
            df = join.run(spark)

            # Build remaining columns again (in case inputs are found in joins)
            df = self.build_columns(
                df, udfs=udfs, raise_exception=i == len(self.joins) - 1
            )

        # Apply filter
        if self.filter:
            df = df.filter(self.filter)

        # Select columns
        cols = []
        if self.selects:
            if isinstance(self.selects, list):
                cols += [F.col(c) for c in self.selects]
            elif isinstance(self.selects, dict):
                cols += [F.col(k).alias(v) for k, v in self.selects.items()]
            df = df.select(cols)

        # Drop columns
        if self.drop_columns:
            logger.info(f"Dropping columns {self.drop_columns}...")
            df = df.drop(*self.drop_columns)

        # Drop duplicates
        pk = self.primary_key
        if self.drop_duplicates and pk:
            logger.info(f"Removing duplicates with {pk}")
            df = df.dropDuplicates([pk])

        return df

    @property
    def apply_changes_kwargs(self):
        cdc = self.source.cdc
        return {
            "apply_as_deletes": cdc.apply_as_deletes,
            "apply_as_truncates": cdc.apply_as_truncates,
            "column_list": cdc.columns,
            "except_column_list": cdc.except_columns,
            "ignore_null_updates": cdc.ignore_null_updates,
            "keys": cdc.primary_keys,
            "sequence_by": cdc.sequence_by,
            "source": self.source.name,
            "stored_as_scd_type": cdc.scd_type,
            "target": self._table.name,
            "track_history_column_list": cdc.track_history_columns,
            "track_history_except_column_list": cdc.track_history_except_columns,
        }



from typing import Union
import pulumi
import pulumi_databricks as databricks
from laktory.resourcesengines.pulumi.base import PulumiResourcesEngine
from laktory.models.compute.notebook import Notebook

from laktory._logger import get_logger

logger = get_logger(__name__)


class PulumiNotebook(PulumiResourcesEngine):
    @property
    def provider(self):
        return "databricks"

    def __init__(
        self,
        name=None,
        notebook: Notebook = None,
        opts=None,
    ):
        if name is None:
            name = f"notebook-{notebook.key}"
        super().__init__(self.t, name, {}, opts)

        opts = pulumi.ResourceOptions(
            parent=self,
            delete_before_replace=True,
        )

        self.notebook = databricks.Notebook(
            f"notebook-{notebook.key}",
            opts=opts,
            **notebook.model_pulumi_dump(),
        )

        access_controls = []
        for permission in notebook.permissions:
            access_controls += [
                databricks.PermissionsAccessControlArgs(
                    permission_level=permission.permission_level,
                    group_name=permission.group_name,
                    service_principal_name=permission.service_principal_name,
                    user_name=permission.user_name,
                )
            ]

        if access_controls:
            self.permissions = databricks.Permissions(
                f"permissions-file-{notebook.key}",
                access_controls=access_controls,
                notebook_path=self.notebook.path,
                opts=opts,
            )



from typing import Literal
from pydantic import Field
from laktory.models.base import BaseModel
from laktory.models.resources import Resources
from laktory.models.permission import Permission


class ClusterAutoScale(BaseModel):
    min_workers: int
    max_workers: int


class ClusterInitScriptVolumes(BaseModel):
    destination: str = None


class ClusterInitScriptWorkspace(BaseModel):
    destination: str = None


class ClusterInitScript(BaseModel):
    volumes: ClusterInitScriptVolumes = None
    workspace: ClusterInitScriptWorkspace = None


class ClusterLibraryCran(BaseModel):
    package: str = None
    repo: str = None


class ClusterLibraryMaven(BaseModel):
    coordinates: str = None
    exclusions: list[str] = None
    repo: str = None


class ClusterLibraryPypi(BaseModel):
    package: str = None
    repo: str = None


class ClusterLibrary(BaseModel):
    cran: ClusterLibraryCran = None
    egg: str = None
    jar: str = None
    maven: ClusterLibraryMaven = None
    pypi: ClusterLibraryPypi = None
    whl: str = None


class Cluster(BaseModel, Resources):
    apply_policy_default_values: bool = None
    autoscale: ClusterAutoScale = None
    autotermination_minutes: int = None
    # aws_attributes
    # azure_attributes
    cluster_id: str = None
    # cluster_log_conf
    # cluster_source
    # cluster_mount_infos
    custom_tags: dict[str, str] = None
    data_security_mode: Literal[
        "NONE", "SINGLE_USER", "USER_ISOLATION"
    ] = "USER_ISOLATION"
    # docker_image
    driver_instance_pool_id: str = None
    driver_node_type_id: str = None
    enable_elastic_disk: bool = None
    enable_local_disk_encryption: bool = None
    # gcp_attributes
    idempotency_token: str = None
    init_scripts: list[ClusterInitScript] = []
    instance_pool_id: str = None
    is_pinned: bool = True
    libraries: list[ClusterLibrary] = []
    name: str = None
    node_type_id: str = Field(...)
    num_workers: int = None
    permissions: list[Permission] = []
    policy_id: str = None
    runtime_engine: Literal["Standard", "Photon"] = None
    single_user_name: str = None
    spark_conf: dict[str, str] = {}
    spark_env_vars: dict[str, str] = {}
    spark_version: str
    ssh_public_keys: list[str] = []
    # workload_type:

    # ----------------------------------------------------------------------- #
    # Resources Engine Methods                                                #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_renames(self):
        return {"name": "cluster_name"}

    @property
    def pulumi_excludes(self) -> list[str]:
        return ["permissions"]

    def deploy_with_pulumi(self, name=None, groups=None, opts=None):
        from laktory.resourcesengines.pulumi.cluster import PulumiCluster

        return PulumiCluster(name=name, cluster=self, opts=opts)



from .compute import *
from .dataevent import DataEvent
from .dataeventheader import DataEventHeader
from .datasources import *
from .grants import *
from .group import Group
from .producer import Producer
from .serviceprincipal import ServicePrincipal
from .secrets import *
from .user import User
from .sql import *



from typing import Union
import pyspark.sql.functions as F
from pyspark.sql.column import Column


COLUMN_OR_NAME = Union[Column, str]
INT_OR_COLUMN = Union[int, COLUMN_OR_NAME]
FLOAT_OR_COLUMN = Union[float, COLUMN_OR_NAME]
STRING_OR_COLUMN = Union[str, Column]


def _col(col: str) -> Column:
    if isinstance(col, Column):
        return col

    return F.col(col)


def _lit(col: str) -> Column:
    if isinstance(col, Column):
        return col

    return F.lit(col)



from typing import Literal
from laktory.models.base import BaseModel

PRIVILEGES = [
    "ALL_PRIVILEGES",
    "READ_VOLUME",
    "WRITE_VOLUME",
]


class VolumeGrant(BaseModel):
    principal: str
    privileges: list[Literal[tuple(PRIVILEGES)]]



from typing import Union

from laktory._logger import get_logger
from laktory.models.base import BaseModel
from laktory.models.datasources.tabledatasource import TableDataSource
from laktory.spark import DataFrame

logger = get_logger(__name__)


class TableJoin(BaseModel):
    left: TableDataSource = None
    other: TableDataSource
    on: list[str]
    how: str = "left"
    time_constraint_interval_lower: str = "60 seconds"
    time_constraint_interval_upper: str = None

    # ----------------------------------------------------------------------- #
    # Readers                                                                 #
    # ----------------------------------------------------------------------- #

    def run(self, spark) -> DataFrame:
        import pyspark.sql.functions as F

        logger.info(f"Executing {self.left.name} {self.how} JOIN {self.other.name}")

        left_df = self.left.read(spark)
        other_df = self.other.read(spark)

        # Add watermark
        other_cols = []
        if self.other.watermark is not None:
            other_cols += [F.col(self.other.watermark.column).alias("_other_wc")]
        other_cols += [F.col(c) for c in other_df.columns]
        other_df = other_df.select(other_cols)

        # Drop duplicates to prevent adding rows to left
        other_df = other_df.dropDuplicates(self.on)

        _join = []
        for c in self.on:
            _join += [f"left.{c} == other.{c}"]
        if self.other.watermark is not None:
            if self.time_constraint_interval_lower:
                _join += [
                    f"left.{self.left.watermark.column} >= other._other_wc - interval {self.time_constraint_interval_lower}"
                ]
            if self.time_constraint_interval_upper:
                _join += [
                    f"left.{self.left.watermark.column} <= other._other_wc + interval {self.time_constraint_interval_upper}"
                ]
        _join = " AND ".join(_join)

        logger.info(f"   ON {_join}")

        logger.info(f"Left Schema:")
        left_df.printSchema()

        logger.info(f"Other Schema:")
        other_df.printSchema()

        df = (
            left_df.alias("left")
            .join(
                other=other_df.alias("other"),
                on=F.expr(_join),
                how=self.how,
            )
            .drop()
        )

        # Drop join columns
        for c in self.on:
            df = df.drop(F.col(f"other.{c}"))
        if self.other.watermark is not None:
            df = df.drop(F.col(f"other._other_wc"))
        logger.info(f"Joined Schema:")
        df.printSchema()

        return df



import pulumi


class PulumiResourcesEngine(pulumi.ComponentResource):
    @property
    def t(self) -> str:
        return f"laktory:{self.provider}:{type(self).__name__.replace('Pulumi', '')}"

    # @property
    # def default_name(self) -> str:
    #     raise NotImplementedError()

    @property
    def provider(self) -> str:
        raise NotImplementedError()

    #
    # def __init__(self, name=None, opts=None, *args, **kwargs):
    #     if name is None:
    #         name = self.default_name
    #     super().__init__(self.t, name, {}, opts)
    #
    #     kwargs["opts"] = kwargs.get("opts", pulumi.ResourceOptions())
    #     kwargs["opts"].parent = self



import json
import pulumi
import pulumi_databricks as databricks

from laktory._logger import get_logger
from laktory._settings import settings
from laktory.models.compute.pipeline import Pipeline
from laktory.resourcesengines.pulumi.base import PulumiResourcesEngine

logger = get_logger(__name__)


class PulumiPipeline(PulumiResourcesEngine):
    @property
    def provider(self):
        return "databricks"

    def __init__(
        self,
        name=None,
        pipeline: Pipeline = None,
        opts=None,
    ):
        if name is None:
            name = f"pipline-{pipeline.name}"
        super().__init__(self.t, name, {}, opts)

        opts = pulumi.ResourceOptions(
            parent=self,
            delete_before_replace=True,
        )

        # ------------------------------------------------------------------- #
        # Pipeline                                                            #
        # ------------------------------------------------------------------- #

        self.pipeline = databricks.Pipeline(
            f"pipline-{pipeline.name}",
            opts=opts,
            **pipeline.model_pulumi_dump(),
        )

        access_controls = []
        for permission in pipeline.permissions:
            access_controls += [
                databricks.PermissionsAccessControlArgs(
                    permission_level=permission.permission_level,
                    group_name=permission.group_name,
                    service_principal_name=permission.service_principal_name,
                    user_name=permission.user_name,
                )
            ]

        if access_controls:
            self.permissions = databricks.Permissions(
                f"permissions-pipeline-{pipeline.name}",
                access_controls=access_controls,
                pipeline_id=self.pipeline.id,
                opts=opts,
            )

        # ------------------------------------------------------------------- #
        # Pipeline Configuration                                              #
        # ------------------------------------------------------------------- #

        source = f"./tmp-{pipeline.name}.json"
        d = pipeline.model_dump(exclude_none=True)
        d = pipeline.inject_vars(d)
        s = json.dumps(d, indent=4)
        with open(source, "w") as fp:
            fp.write(s)
        filepath = f"{settings.workspace_laktory_root}pipelines/{pipeline.name}.json"
        self.conf = databricks.WorkspaceFile(
            f"file-{filepath}",
            path=filepath,
            # md5=hashlib.md5(s.encode('utf-8')).hexdigest(),
            source=source,
            opts=opts,
        )
        # os.remove(filepath)

        if access_controls:
            self.conf_permissions = databricks.Permissions(
                f"permissions-file-{filepath}",
                access_controls=[
                    databricks.PermissionsAccessControlArgs(
                        permission_level="CAN_READ",
                        group_name="account users",
                    )
                ],
                workspace_file_path=filepath,
                opts=opts,
            )



from datetime import datetime

from laktory.models import WorkspaceFile


def test_workspace_file():
    workspace_file = WorkspaceFile(
        source="../libraries/init_scripts/install_laktory.sh",
        dirpath="/init_scripts/",
        permissions=[{"permission_level": "CAN_READ", "group_name": "account users"}],
    )
    assert workspace_file.filename == "install_laktory.sh"
    assert workspace_file.key == "init_scripts-install_laktory"
    assert workspace_file.path == "/init_scripts/install_laktory.sh"

    assert workspace_file.permissions[0].permission_level == "CAN_READ"
    assert workspace_file.permissions[0].group_name == "account users"


if __name__ == "__main__":
    test_workspace_file()



# def _dict_to_sql(d, mode="data"):
#     if mode == "data":
#         statement = "named_struct("
#         for key, value in d.items():
#             statement += f"'{key}', {py_to_sql(value, mode=mode)}, "
#         statement = statement.rstrip(", ") + ")"
#
#     elif mode == "schema":
#         statement = "STRUCT<"
#         for key, value in d.items():
#             statement += f"{key}: {py_to_sql(value, mode=mode)}, "
#         statement = statement.rstrip(", ") + ">"
#
#     return statement
#
#
# def _list_to_sql(l, mode="data"):
#     if mode == "data":
#         statement = "ARRAY("
#         for value in l:
#             statement += f"{py_to_sql(value, mode=mode)}, "
#         statement = statement.rstrip(", ") + ")"
#     elif mode == "schema":
#         statement = "ARRAY<"
#         for value in l:
#             statement += f"{py_to_sql(value, mode=mode)}, "
#         statement = statement.rstrip(", ") + ">"
#
#     return statement
#
#
# def py_to_sql(value, mode="data"):
#     if isinstance(value, str):
#         if mode == "data":
#             return f"'{value}'"
#         elif mode == "schema":
#             return f"{value}"
#     elif value is None:
#         return f"null"
#     elif isinstance(value, list):
#         return _list_to_sql(value, mode=mode)
#     elif isinstance(value, dict):
#         return _dict_to_sql(value, mode=mode)
#     else:
#         return f"{value}"
#



import os
import json
from datetime import datetime
from zoneinfo import ZoneInfo
from typing import Any
from typing import Literal
from typing import Union

from pydantic import Field
from pydantic import ConfigDict

from laktory.models.dataeventheader import DataEventHeader
from laktory.models.producer import Producer
from laktory._settings import settings
from laktory._logger import get_logger

logger = get_logger(__name__)

EXCLUDES = [
    "events_root",
    "event_path",
    "tstamp_col",
    "tstamp_in_path",
]


class DataEvent(DataEventHeader):
    model_config = ConfigDict(populate_by_name=True)
    name: str = Field(..., alias="event_name")
    description: Union[str, None] = Field(None, alias="event_description")
    producer: Producer = Field(None, alias="event_producer")
    data: dict
    tstamp_col: str = "created_at"
    tstamp_in_path: bool = True

    def model_post_init(self, __context):
        # Add metadata
        self.data["_name"] = self.name
        self.data["_producer_name"] = self.producer.name
        tstamp = self.data.get(self.tstamp_col)
        if isinstance(tstamp, str):
            tstamp = datetime.fromisoformat(tstamp)
        elif tstamp is None:
            tstamp = datetime.utcnow()
        if not tstamp.tzinfo:
            tstamp = tstamp.replace(tzinfo=ZoneInfo("UTC"))
        self.data["_created_at"] = tstamp

    # ----------------------------------------------------------------------- #
    # Properties                                                              #
    # ----------------------------------------------------------------------- #

    @property
    def created_at(self) -> datetime:
        return self.data["_created_at"]

    # ----------------------------------------------------------------------- #
    # Paths                                                                   #
    # ----------------------------------------------------------------------- #

    @property
    def dirpath(self) -> str:
        dirpath = self.event_root
        if self.tstamp_in_path:
            t = self.created_at
            dirpath += f"{t.year:04d}/{t.month:02d}/{t.day:02d}/"
        return dirpath

    def get_filename(self, fmt: str = "json", suffix: str = None) -> str:
        filename = self.name
        if suffix is not None:
            filename += f"_{suffix}"
        if self.tstamp_in_path:
            t = self.created_at
            const = {
                "mus": {"s": 1e-6},
                "s": {"ms": 1000},
            }  # TODO: replace with constants
            total_ms = int(
                (t.second + t.microsecond * const["mus"]["s"]) * const["s"]["ms"]
            )
            time_str = f"{t.hour:02d}{t.minute:02d}{total_ms:05d}Z"
            filename += f"_{t.year:04d}{t.month:02d}{t.day:02d}T{time_str}"

        filename += f".{fmt}"

        return filename

    def get_landing_filepath(self, fmt="json", suffix=None):
        return os.path.join(self.dirpath, self.get_filename(fmt, suffix))

    def get_storage_filepath(self, fmt="json", suffix=None):
        path = self.get_landing_filepath(fmt=fmt, suffix=suffix)
        path = path.replace(settings.workspace_landing_root, "/")
        return path

    # ----------------------------------------------------------------------- #
    # Output                                                                  #
    # ----------------------------------------------------------------------- #

    def model_dump(self, *args, **kwargs) -> dict[str, Any]:
        kwargs["exclude"] = kwargs.pop("exclude", EXCLUDES)
        kwargs["by_alias"] = kwargs.pop("by_alias", True)
        kwargs["mode"] = kwargs.pop("mode", "json")
        return super().model_dump(*args, **kwargs)

    def model_dump_json(self, *args, **kwargs) -> str:
        kwargs["exclude"] = kwargs.pop("exclude", EXCLUDES)
        return super().model_dump_json(*args, **kwargs)

    def _overwrite_or_skip(
        self, f: callable, path: str, exists: bool, overwrite: bool, skip: bool
    ):
        if exists:
            if skip:
                logger.info(f"Event {self.name} ({path}) already exists. Skipping.")

            else:
                if overwrite:
                    logger.info(f"Writing event {self.name} to {path}")
                    f()
                else:
                    raise FileExistsError(f"Object {path} already exists.")

        else:
            logger.info(f"Writing event {self.name} to {path}")
            f()

    def to_path(
        self,
        suffix: str = None,
        fmt: str = "json",
        overwrite: bool = False,
        skip_if_exists: bool = False,
    ) -> None:
        path = self.get_landing_filepath(suffix=suffix, fmt=fmt)
        dirpath = os.path.dirname(path)

        if not os.path.exists(dirpath):
            os.makedirs(dirpath)

        def _write():
            if fmt != "json":
                raise NotImplementedError()

            with open(path, "w") as fp:
                fp.write(self.model_dump_json())

        self._overwrite_or_skip(
            f=_write,
            path=path,
            exists=os.path.exists(path),
            overwrite=overwrite,
            skip=skip_if_exists,
        )

    def to_azure_storage_container(
        self,
        suffix: str = None,
        fmt: str = "json",
        container_client: Any = None,
        account_url: str = None,
        container_name: str = "landing",
        overwrite: bool = False,
        skip_if_exists: bool = False,
    ) -> None:
        # Set container client
        if container_client is None:
            from azure.storage.blob import ContainerClient
            from azure.identity import DefaultAzureCredential

            if account_url:
                # From account URL
                # TODO: test
                container_client = ContainerClient(
                    account_url,
                    credential=DefaultAzureCredential(),
                    container_name=container_name,
                )

            elif settings.lakehouse_sa_conn_str is not None:
                # From connection string
                container_client = ContainerClient.from_connection_string(
                    conn_str=settings.lakehouse_sa_conn_str,
                    container_name=container_name,
                )

            else:
                raise ValueError(
                    "Provide a valid container client, an account url or a connection string in LAKEHOUSE_SA_CONN_STR"
                )

        path = self.get_storage_filepath(suffix=suffix, fmt=fmt)
        blob = container_client.get_blob_client(path)

        def _write():
            if fmt != "json":
                raise NotImplementedError()

            blob.upload_blob(
                self.model_dump_json(),
                overwrite=overwrite,
            )

        self._overwrite_or_skip(
            f=_write,
            path=path,
            exists=blob.exists(),
            overwrite=overwrite,
            skip=skip_if_exists,
        )

    def to_aws_s3_bucket(
        self,
        bucket_name,
        suffix: str = None,
        fmt: str = "json",
        s3_resource: Any = None,
        overwrite: bool = False,
        skip_if_exists: bool = False,
    ) -> None:
        import boto3

        if s3_resource is None:
            s3_resource = boto3.resource(
                service_name="s3",
                aws_access_key_id=settings.aws_access_key_id,
                aws_secret_access_key=settings.aws_secret_access_key,
                region_name=settings.aws_region,
            )

        bucket = s3_resource.Bucket(bucket_name)

        path = self.get_storage_filepath(suffix=suffix, fmt=fmt)[1:]

        object_exists = True
        try:
            o = bucket.Object(path)
            o.load()
        except Exception as e:
            if "404" in str(e):
                object_exists = False
            else:
                raise e

        def _write():
            if fmt != "json":
                raise NotImplementedError()

            bucket.put_object(
                Key=path,
                Body=self.model_dump_json(),
            )

        self._overwrite_or_skip(
            f=_write,
            path=path,
            exists=object_exists,
            overwrite=overwrite,
            skip=skip_if_exists,
        )

    def to_gcp_storage_bucket(self):
        raise NotImplementedError()

    def to_databricks(
        self,
        suffix: str = None,
        fmt: str = "json",
        overwrite: bool = False,
        skip_if_exists: bool = False,
        storage_type: Literal["VOLUME", "MOUNT"] = "VOLUME",
    ) -> None:
        path = self.get_landing_filepath(suffix=suffix, fmt=fmt)
        if storage_type == "MOUNT":
            path = "/dbfs" + path
        dirpath = os.path.dirname(path)

        if not os.path.exists(dirpath):
            os.makedirs(dirpath)

        def _write():
            if fmt != "json":
                raise NotImplementedError()

            with open(path, "w") as fp:
                fp.write(self.model_dump_json())

        self._overwrite_or_skip(
            f=_write,
            path=path,
            exists=os.path.exists(path),
            overwrite=overwrite,
            skip=skip_if_exists,
        )



import time

# from databricks.sdk import WorkspaceClient as _WorkspaceClient
# from databricks.sdk.service.sql import StatementState

from laktory import settings


class WorkspaceClient:
    pass
    # TODO: Implement


#
# class WorkspaceClient(_WorkspaceClient):
#     def __init__(self, *args, **kwargs):
#         kwargs["host"] = kwargs.get("host", settings.databricks_host)
#         kwargs["token"] = kwargs.get("token", settings.databricks_token)
#         super().__init__(*args, **kwargs)
#
#     def execute_statement_and_wait(
#         self, statement, warehouse_id=None, catalog_name=None
#     ):
#         # Settings
#         if warehouse_id is None:
#             warehouse_id = settings.databricks_warehouse_id
#
#         r = self.statement_execution.execute_statement(
#             statement=statement,
#             catalog=catalog_name,
#             warehouse_id=warehouse_id,
#         )
#         statement_id = r.statement_id
#         state = r.status.state
#
#         while state in [StatementState.PENDING, StatementState.RUNNING]:
#             r = self.statement_execution.get_statement(statement_id)
#             time.sleep(1)
#             state = r.status.state
#
#         if state != StatementState.SUCCEEDED:
#             # TODO: Create specific error
#             print(r.status.error.message)
#             raise Exception(r.status.error)
#
#         return r



from laktory.resourcesengines.pulumi.base import PulumiResourcesEngine
from laktory.resourcesengines.pulumi.catalog import PulumiCatalog
from laktory.resourcesengines.pulumi.cluster import PulumiCluster
from laktory.resourcesengines.pulumi.group import PulumiGroup
from laktory.resourcesengines.pulumi.notebook import PulumiNotebook
from laktory.resourcesengines.pulumi.pipeline import PulumiPipeline
from laktory.resourcesengines.pulumi.schema import PulumiSchema
from laktory.resourcesengines.pulumi.serviceprincipal import PulumiServicePrincipal
from laktory.resourcesengines.pulumi.user import PulumiUser
from laktory.resourcesengines.pulumi.volume import PulumiVolume



from laktory.workspaceclient import WorkspaceClient


class DatabricksResourcesEngine:
    def __init__(self):
        pass

    @property
    def workspace_client(self):
        return WorkspaceClient()

    # @classmethod
    # def model_sql_schema(cls, types: dict = None):
    #     if types is None:
    #         types = cls.model_serialized_types()
    #
    #     schema = (
    #         "("
    #         + ", ".join(f"{k} {py_to_sql(v, mode='schema')}" for k, v in types.items())
    #         + ")"
    #     )
    #
    #     if "<>" in schema:
    #         raise ValueError(
    #             f"Some types are undefined sql schema can't be defined\n{schema}"
    #         )
    #
    #     if "null" in schema:
    #         raise ValueError(
    #             f"Some types are undefined sql schema can't be defined\n{schema}"
    #         )
    #
    #     return schema



from typing import Union

import pulumi
import pulumi_databricks as databricks

from laktory.resourcesengines.pulumi.base import PulumiResourcesEngine
from laktory.models.group import Group
from laktory.models.serviceprincipal import ServicePrincipal

from laktory._logger import get_logger

logger = get_logger(__name__)


class PulumiServicePrincipal(PulumiResourcesEngine):
    @property
    def provider(self):
        return "databricks"

    def __init__(
        self,
        name=None,
        service_principal: ServicePrincipal = None,
        group_ids: dict[str, str] = None,
        opts=None,
    ):
        sp = service_principal
        if name is None:
            name = f"service-principal-{sp.display_name}"
        super().__init__(self.t, name, {}, opts)

        opts = pulumi.ResourceOptions(
            parent=self,
            delete_before_replace=True,
        )

        self.sp = databricks.ServicePrincipal(
            f"service-principal-{sp.display_name}",
            opts=opts,
            **sp.model_pulumi_dump(),
        )

        self.roles = []
        for role in sp.roles:
            self.roles += [
                databricks.ServicePrincipalRole(
                    f"service-principal-role-{sp.display_name}-{role}",
                    service_principal_id=self.sp.id,
                    role=role,
                    opts=opts,
                )
            ]

        if not group_ids:
            if sp.groups:
                logger.warning(
                    "User is member of groups, but groups have not been provided. Group member resources will "
                    "be skipped."
                )
            return

        # Group Member
        self.group_members = []
        for g in sp.groups:
            # Find matching group
            group_id = group_ids.get(g, None)

            if group_id:
                self.group_members += [
                    databricks.GroupMember(
                        f"group-member-{sp.display_name}-{g}",
                        group_id=group_id,
                        member_id=self.sp.id,
                        opts=opts,
                    )
                ]



from laktory.models.base import BaseModel
from laktory.models.resources import Resources


class ServicePrincipal(BaseModel, Resources):
    display_name: str
    application_id: str = None
    allow_cluster_create: bool = False
    groups: list[str] = []
    roles: list[str] = []

    # Deployment options
    disable_as_user_deletion: bool = False

    # ----------------------------------------------------------------------- #
    # Resources Engine Methods                                                #
    # ----------------------------------------------------------------------- #
    @property
    def pulumi_excludes(self) -> list[str]:
        return ["groups", "roles"]

    def deploy_with_pulumi(self, name=None, group_ids=None, opts=None):
        from laktory.resourcesengines.pulumi.serviceprincipal import (
            PulumiServicePrincipal,
        )

        return PulumiServicePrincipal(
            name=name, service_principal=self, group_ids=group_ids, opts=opts
        )



from laktory.spark import DataFrame
from typing import Union
from typing import Literal
from typing import Any

from laktory.models.base import BaseModel
from laktory.models.datasources.basedatasource import BaseDataSource
from laktory._logger import get_logger

logger = get_logger(__name__)


class Watermark(BaseModel):
    column: str
    threshold: str


class TableDataSourceCDC(BaseModel):
    apply_as_deletes: Union[str, None] = None
    apply_as_truncates: Union[str, None] = None
    columns: Union[list[str], None] = []
    except_columns: Union[list[str], None] = []
    ignore_null_updates: Union[bool, None] = None
    primary_keys: list[str]
    scd_type: Literal[1, 2] = None
    sequence_by: str
    track_history_columns: Union[list[str], None] = None
    track_history_except_columns: Union[list[str], None] = None


class TableDataSource(BaseDataSource):
    _df: Any = None
    catalog_name: Union[str, None] = None
    cdc: Union[TableDataSourceCDC, None] = None
    selects: Union[list[str], dict[str, str], None] = None
    filter: Union[str, None] = None
    from_pipeline: Union[bool, None] = True
    name: Union[str, None]
    schema_name: Union[str, None] = None
    watermark: Union[Watermark, None] = None

    # ----------------------------------------------------------------------- #
    # Properties                                                              #
    # ----------------------------------------------------------------------- #

    @property
    def full_name(self) -> str:
        name = ""
        if self.catalog_name is not None:
            name = self.catalog_name

        if self.schema_name is not None:
            if name == "":
                name = self.schema_name
            else:
                name += f".{self.schema_name}"

        if name == "":
            name = self.name
        else:
            name += f".{self.name}"

        return name

    # ----------------------------------------------------------------------- #
    # Readers                                                                 #
    # ----------------------------------------------------------------------- #

    def _read(self, spark) -> DataFrame:
        from laktory.dlt import read
        from laktory.dlt import read_stream

        if self._df is not None:
            logger.info(f"Reading {self.full_name} from memory")
            df = self._df
        elif self.read_as_stream:
            logger.info(f"Reading {self.full_name} as stream")
            if self.from_pipeline:
                df = read_stream(self.full_name)
            else:
                df = spark.readStream.format("delta").table(self.full_name)
        else:
            logger.info(f"Reading {self.full_name} as static")
            if self.from_pipeline:
                df = read(self.full_name)
            else:
                df = spark.read.table(self.full_name)

        return df

    def read(self, spark) -> DataFrame:
        import pyspark.sql.functions as F

        df = self._read(spark)

        # Apply filter
        if self.filter:
            df = df.filter(self.filter)

        # Columns
        cols = []
        if self.selects:
            if isinstance(self.selects, list):
                cols += [F.col(c) for c in self.selects]
            elif isinstance(self.selects, dict):
                cols += [F.col(k).alias(v) for k, v in self.selects.items()]
            df = df.select(cols)

        # Apply Watermark
        if self.watermark:
            df = df.withWatermark(
                self.watermark.column,
                self.watermark.threshold,
            )

        return df



from typing import Literal
from laktory.models.base import BaseModel
from laktory.models.resources import Resources
from laktory.models.permission import Permission


class WarehouseCustomTag(BaseModel):
    key: str
    value: str


class WarehouseTags(BaseModel):
    custom_tags: list[WarehouseCustomTag] = []


class Warehouse(BaseModel, Resources):
    cluster_size: Literal[
        "2X-Small",
        "X-Small",
        "Small",
        "Medium",
        "Large",
        "X-Large",
        "2X-Large",
        "3X-Large",
        "4X-Large",
    ]
    auto_stop_mins: int = None
    channel_name: Literal["CHANNEL_NAME_CURRENT", "CHANNEL_NAME_PREVIEW"] = None
    # data_source_id
    enable_photon: bool = None
    enable_serverless_compute: bool = None
    instance_profile_arn: str = None
    jdbc_url: str = None
    max_num_clusters: int = None
    min_num_clusters: int = None
    name: str = None
    num_clusters: int = None
    # odbc_params
    spot_instance_policy: Literal["COST_OPTIMIZED", "RELIABILITY_OPTIMIZED"] = None
    # state
    tags: WarehouseTags = None
    warehouse_type: Literal["CLASSIC", "PRO"] = None

    # Deployment Options
    permissions: list[Permission] = []

    # ----------------------------------------------------------------------- #
    # Resources Engine Methods                                                #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_excludes(self) -> list[str]:
        return ["permissions"]

    def model_pulumi_dump(self, *args, **kwargs):
        d = super().model_pulumi_dump(*args, **kwargs)
        d["channel"] = {"name": d.pop("channel_name")}
        return d

    def deploy_with_pulumi(self, name=None, groups=None, opts=None):
        from laktory.resourcesengines.pulumi.warehouse import PulumiWarehouse

        return PulumiWarehouse(name=name, warehouse=self, opts=opts)



from typing import Any
from typing import Literal
from pydantic import model_validator
from pydantic import Field
from laktory.models.base import BaseModel
from laktory.models.resources import Resources
from laktory.models.permission import Permission
from laktory.models.compute.cluster import Cluster
from laktory.models.compute.cluster import ClusterLibrary


class JobCluster(Cluster):
    is_pinned: bool = Field(None)
    libraries: list[Any] = Field(None)
    permissions: list[Any] = Field(None)

    @model_validator(mode="after")
    def excluded_fields(self) -> Any:
        for f in [
            "is_pinned",
            "libraries",
            "permissions",
        ]:
            if getattr(self, f, None) not in [None, [], {}]:
                raise ValueError(f"Field {f} should be null")

        return self


class JobContinuous(BaseModel):
    pause_status: Literal["PAUSED", "UNPAUSED"] = None


class JobEmailNotifications(BaseModel):
    alert_on_last_attempt: bool = None
    no_alert_for_skipped_runs: bool = None
    on_duration_warning_threshold_exceededs: list[str] = None
    on_failures: list[str] = None
    on_starts: list[str] = None
    on_success: list[str] = None


class JobHealthRule(BaseModel):
    metric: str = None
    op: str = None
    value: int = None


class JobHealth(BaseModel):
    rules: list[JobHealthRule] = None


class JobNotificationSettings(BaseModel):
    no_alert_for_canceled_runs: bool = None
    no_alert_for_skipped_runs: bool = None


class JobParameter(BaseModel):
    default: str = None
    name: str = None


class JobRunAs(BaseModel):
    service_principal_name: str = None
    user_name: str = None


class JobSchedule(BaseModel):
    quartz_cron_expression: str
    timezone_id: str
    pause_status: str


class JobTaskConditionTask(BaseModel):
    left: str = None
    op: str = None
    right: str = None


class JobTaskDependsOn(BaseModel):
    task_key: str = None
    outcome: str = None


class JobTaskNotebookTask(BaseModel):
    notebook_path: str = None
    base_parameters: dict[str, Any] = None
    source: str = None


class JobTaskPipelineTask(BaseModel):
    pipeline_id: str = None
    full_refresh: bool = None


class JobTaskRunJobTask(BaseModel):
    job_id: str = None
    job_parameters: dict[str, Any]


class JobTaskSqlTaskQuery(BaseModel):
    query_id: str = None


class JobTaskSqlTaskAlertSubscription(BaseModel):
    destination_id: str = None
    user_name: str = None


class JobTaskSQLTaskAlert(BaseModel):
    alert_id: str = None
    subscriptions: list[JobTaskSqlTaskAlertSubscription] = None
    pause_subscriptions: bool = None


class JobTaskSqlTaskDashboard(BaseModel):
    dashboard_id: str = None
    custom_subject: list[JobTaskSqlTaskAlertSubscription] = None
    subscriptions: list[JobTaskSqlTaskAlertSubscription] = None


class JobTaskSqlTaskFile(BaseModel):
    path: str = None


class JobTaskSQLTask(BaseModel):
    alert: JobTaskSQLTaskAlert = None
    dashboard: JobTaskSqlTaskDashboard = None
    file: JobTaskSqlTaskFile = None
    parameters: dict[str, Any] = None
    query: JobTaskSqlTaskQuery
    warehouse_id: str = None


class JobTask(BaseModel):
    # compute_key: str = None
    condition_task: JobTaskConditionTask = None
    depends_ons: list[JobTaskDependsOn] = None
    description: str = None
    email_notifications: JobEmailNotifications = None
    existing_cluster_id: str = None
    health: JobHealth = None
    job_cluster_key: str = None
    libraries: list[ClusterLibrary] = None
    max_retries: int = None
    min_retry_interval_millis: int = None
    # new_cluster: Cluster = None
    notebook_task: JobTaskNotebookTask = None
    notification_settings: JobNotificationSettings = None
    pipeline_task: JobTaskPipelineTask = None
    # python_wheel_task:
    retry_on_timeout: bool = None
    run_if: str = None
    run_job_task: JobTaskRunJobTask = None
    # spark_jar_task:
    # spark_python_task:
    sql_task: JobTaskSQLTask = None
    task_key: str = None
    timeout_seconds: int = None


class JobTriggerFileArrival(BaseModel):
    url: str = None
    min_time_between_triggers_seconds: int = None
    wait_after_last_change_seconds: int = None


class JobTrigger(BaseModel):
    file_arrival: JobTriggerFileArrival
    pause_status: Literal["PAUSED", "UNPAUSED"] = None


class JobWebhookNotificationsOnDurationWarningThresholdExceeded(BaseModel):
    id: str = None


class JobWebhookNotificationsOnFailure(BaseModel):
    id: str = None


class JobWebhookNotificationsOnStart(BaseModel):
    id: str = None


class JobWebhookNotificationsOnSuccess(BaseModel):
    id: str = None


class JobWebhookNotifications(BaseModel):
    on_duration_warning_threshold_exceededs: list[
        JobWebhookNotificationsOnDurationWarningThresholdExceeded
    ] = None
    on_failures: list[JobWebhookNotificationsOnFailure] = None
    on_starts: list[JobWebhookNotificationsOnStart] = None
    on_successes: list[JobWebhookNotificationsOnSuccess] = None


class Job(BaseModel, Resources):
    clusters: list[JobCluster] = []
    continuous: JobContinuous = None
    control_run_state: bool = None
    email_notifications: JobEmailNotifications = None
    format: str = None
    health: JobHealth = None
    max_concurrent_runs: int = None
    max_retries: int = None
    min_retry_interval_millis: int = None
    name: str = None
    notification_settings: JobNotificationSettings = None
    parameters: list[JobParameter] = []
    permissions: list[Permission] = []
    # queue: Optional[JobQueueArgs] = None
    retry_on_timeout: bool = None
    run_as: JobRunAs = None
    schedule: JobSchedule = None
    tags: dict[str, Any] = {}
    tasks: list[JobTask] = []
    timeout_seconds: int = None
    trigger: JobTrigger = None
    webhook_notifications: JobWebhookNotifications = None

    # ----------------------------------------------------------------------- #
    # Resources Engine Methods                                                #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_excludes(self) -> list[str]:
        return ["permissions"]

    @property
    def pulumi_renames(self) -> dict[str, str]:
        return {"clusters": "job_clusters"}

    def model_pulumi_dump(self, *args, **kwargs):
        d = super().model_pulumi_dump(*args, **kwargs)
        _clusters = []
        for c in d.get("job_clusters", []):
            name = c.pop("name")
            _clusters += [
                {
                    "job_cluster_key": name,
                    "new_cluster": c,
                }
            ]
        d["job_clusters"] = _clusters

        return d

    def deploy_with_pulumi(self, name=None, opts=None):
        from laktory.resourcesengines.pulumi.job import PulumiJob

        return PulumiJob(name=name, job=self, opts=opts)



from laktory.models.base import BaseModel
from laktory.models.resources import Resources
from laktory.models.grants.volumegrant import VolumeGrant


class Volume(BaseModel, Resources):
    name: str
    catalog_name: str = None
    schema_name: str = None
    volume_type: str = "MANAGED"
    storage_location: str = None
    grants: list[VolumeGrant] = None

    # ----------------------------------------------------------------------- #
    # Computed fields                                                         #
    # ----------------------------------------------------------------------- #

    @property
    def parent_full_name(self) -> str:
        _id = ""
        if self.catalog_name:
            _id += self.catalog_name

        if self.schema_name:
            if _id == "":
                _id = self.schema_name
            else:
                _id += f".{self.schema_name}"

        return _id

    @property
    def full_name(self) -> str:
        _id = self.name
        if self.parent_full_name is not None:
            _id = f"{self.parent_full_name}.{_id}"
        return _id

    # ----------------------------------------------------------------------- #
    # Resources Engine Methods                                                #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_excludes(self) -> list[str]:
        return ["grants"]

    def deploy_with_pulumi(self, name=None, **kwargs):
        from laktory.resourcesengines.pulumi.volume import PulumiVolume

        return PulumiVolume(name=name, volume=self, **kwargs)



from typing import Union

import pulumi
import pulumi_databricks as databricks
from laktory.resourcesengines.pulumi.base import PulumiResourcesEngine
from laktory.models.sql.volume import Volume

from laktory._logger import get_logger

logger = get_logger(__name__)


class PulumiVolume(PulumiResourcesEngine):
    @property
    def provider(self):
        return "databricks"

    def __init__(
        self,
        name=None,
        volume: Volume = None,
        opts=None,
    ):
        if name is None:
            name = f"volume-{volume.full_name}"
        super().__init__(self.t, name, {}, opts)

        opts = pulumi.ResourceOptions(
            parent=self,
        )

        # Volume
        self.volume = databricks.Volume(
            f"volume-{volume.full_name}",
            opts=opts,
            **volume.model_pulumi_dump(),
        )

        # Volume grants
        _opts = opts.merge(pulumi.ResourceOptions(depends_on=self.volume))
        if volume.grants:
            self.grants = databricks.Grants(
                f"grants-{volume.full_name}",
                volume=volume.full_name,
                grants=[
                    databricks.GrantsGrantArgs(
                        principal=g.principal, privileges=g.privileges
                    )
                    for g in volume.grants
                ],
                opts=_opts,
            )



from typing import Literal
from laktory.models.base import BaseModel
from laktory.models.grants.tablegrant import PRIVILEGES as TABLE_PRIVILEGES
from laktory.models.grants.volumegrant import PRIVILEGES as VOLUME_PRIVILEGES
from laktory.models.grants.functiongrant import PRIVILEGES as FUNCTION_PRIVILEGES
from laktory.models.grants.registeredmodelgrant import PRIVILEGES as MODEL_PRIVILEGES
from laktory.models.grants.viewgrant import PRIVILEGES as VIEWL_PRIVILEGES


PRIVILEGES = (
    TABLE_PRIVILEGES
    + VOLUME_PRIVILEGES
    + FUNCTION_PRIVILEGES
    + MODEL_PRIVILEGES
    + VIEWL_PRIVILEGES
    + [
        "ALL_PRIVILEGES",
        "CREATE_FUNCTION",
        "CREATE_MATERIALIZED_VIEW",
        "CREATE_MODEL",
        "CREATE_TABLE",
        "CREATE_VOLUME",
        "USE_SCHEMA",
    ]
)


class SchemaGrant(BaseModel):
    principal: str
    privileges: list[Literal[tuple(PRIVILEGES)]]



from typing import Literal
from typing import Any
from pydantic import Field
from pydantic import model_validator
from laktory.models.base import BaseModel
from laktory.models.resources import Resources
from laktory.models.secrets.secret import Secret


class SecretScopePermission(BaseModel):
    permission: Literal["READ", "WRITE", "MANAGE"] = None
    principal: str = None
    scope: str = None


class SecretScopeKeyvaultMetadata(BaseModel):
    dns_name: str = None
    resource_id: str = None


class SecretScope(BaseModel, Resources):
    backend_type: Literal["DATABRICKS", "AZURE_KEYVAULT"] = "DATABRICKS"
    keyvault_metadata: SecretScopeKeyvaultMetadata = None
    name: str = Field(...)
    secrets: list[Secret] = []
    permissions: list[SecretScopePermission] = []

    @model_validator(mode="after")
    def set_secrets_scope(self) -> Any:
        for s in self.secrets:
            s.scope = self.name

        return self

    # ----------------------------------------------------------------------- #
    # Resources Engine Methods                                                #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_excludes(self) -> list[str]:
        return ["permissions", "secrets"]

    def deploy_with_pulumi(self, name=None, groups=None, opts=None):
        from laktory.resourcesengines.pulumi.secretscope import PulumiSecretScope

        return PulumiSecretScope(name=name, secret_scope=self, opts=opts)



from typing import Literal
from laktory.models.base import BaseModel

PRIVILEGES = [
    "ALL_PRIVILEGES",
    "CREATE_EXTERNAL_TABLE",
    "CREATE_EXTERNAL_VOLUME",
    "READ_FILES",
    "WRITE_FILES",
    "CREATE_MANAGED_STORAGE",
]


class ExternalLocationGrant(BaseModel):
    principal: str
    privileges: list[Literal[tuple(PRIVILEGES)]]



from pydantic import model_validator
from laktory.spark import DataFrame
from typing import Literal
from typing import Any

from laktory.models.dataeventheader import DataEventHeader
from laktory.models.datasources.basedatasource import BaseDataSource
from laktory._logger import get_logger

logger = get_logger(__name__)

TYPES = (
    "STORAGE_DUMPS",
    "STORAGE_EVENTS",
    "STREAM_KINESIS",
    "STREAM_KAFKA",
)

FORMATS = (
    "JSON",
    "CSV",
    "PARQUET",
)


class EventDataSource(BaseDataSource, DataEventHeader):
    type: Literal[TYPES] = "STORAGE_EVENTS"
    fmt: Literal[FORMATS] = "JSON"
    multiline: bool = False
    header: bool = True
    read_options: dict[str, str] = {}

    # ----------------------------------------------------------------------- #
    # Readers                                                                 #
    # ----------------------------------------------------------------------- #

    def _read_storage(self, spark) -> DataFrame:
        if self.read_as_stream:
            logger.info(f"Reading {self.event_root} as stream")

            # Set reader
            reader = (
                spark.readStream.format("cloudFiles")
                .option("cloudFiles.format", self.fmt)
                .option("cloudFiles.schemaLocation", self.event_root)
                .option("cloudFiles.inferColumnTypes", True)
                .option("cloudFiles.schemaEvolutionMode", "addNewColumns")
                .option("cloudFiles.allowOverwrites", True)
            )

        else:
            logger.info(f"Reading {self.event_root} as static")

            # Set reader
            reader = spark.read.format(self.fmt)

        reader = (
            reader.option("multiLine", self.multiline)  # only apply to JSON format
            .option("mergeSchema", True)
            .option("recursiveFileLookup", True)
            .option("header", self.header)  # only apply to CSV format
        )
        if self.read_options:
            reader = reader.options(**self.read_options)

        # Load
        df = reader.load(self.event_root)

        # Not supported by UC
        # .withColumn("file", F.input_file_name())

        return df

    def read(self, spark) -> DataFrame:
        if self.type == "STORAGE_EVENTS":
            return self._read_storage(spark)
        else:
            raise NotImplementedError()



from typing import Any
from pydantic import model_validator
from laktory.models.base import BaseModel


class Permission(BaseModel):
    permission_level: str
    user_name: str = None
    group_name: str = None
    service_principal_name: str = None

    @model_validator(mode="after")
    def singlue_input(self) -> Any:
        count = 0
        if self.user_name:
            count += 1
        if self.group_name:
            count += 1
        if self.service_principal_name:
            count += 1

        if count == 0:
            raise ValueError(
                "At least one of `user_name`, `group_name`, `service_principal_name` must be specified."
            )
        if count > 1:
            raise ValueError(
                "Ony one of `user_name`, `group_name`, `service_principal_name` must be specified."
            )

        return self



from abc import abstractmethod
from typing import Union
from laktory.spark import DataFrame

from laktory.models.base import BaseModel


class BaseDataSource(BaseModel):
    read_as_stream: Union[bool, None] = True

    @abstractmethod
    def read(self, spark) -> DataFrame:
        raise NotImplementedError()

    @property
    def is_cdc(self):
        return getattr(self, "cdc", None) is not None



import os

from laktory.models import Notebook

root_dir = os.path.dirname(__file__)


def test_notebook():
    nb0 = Notebook(source="./hello_world.py")
    nb1 = Notebook(source="./notebooks/demos/hello_world.py")
    nb2 = Notebook(source="./notebooks/demos/hello_world.py", dirpath="/files/")

    assert nb0.path is None
    assert nb1.path == "/demos/hello_world.py"
    assert nb2.path == "/files/hello_world.py"


if __name__ == "__main__":
    test_notebook()



from pydantic import field_validator
from pydantic import model_validator
from typing import Union
from typing import Any
from typing import Callable

from laktory._logger import get_logger
from laktory.models.base import BaseModel
from laktory.spark import DataFrame
from laktory.spark import Column as SparkColumn
from laktory.models.sql.column import Column

logger = get_logger(__name__)


class Window(BaseModel):
    time_column: str
    window_duration: str
    slide_duration: Union[str, None] = None
    start_time: Union[str, None] = None


# TODO: Add support for partition by and drop
"""
w = Window.partitionBy("account_id").orderBy(F.col("created_at").desc())
df = df.withColumn("_row", F.row_number().over(w)).filter(F.col("_row") == 1)
df = df.drop("_row")
"""


class TableAggregation(BaseModel):
    groupby_window: Union[Window, None] = None
    groupby_columns: Union[list[str], None] = []
    agg_expressions: list[Column] = []

    @model_validator(mode="after")
    def groupby(self) -> Any:
        if self.groupby_window is None and len(self.groupby_columns) == 0:
            raise ValueError(
                "Either `groupby_window` or `groupby_columns` must be specified."
            )

        return self

    # ----------------------------------------------------------------------- #
    # Methods                                                                 #
    # ----------------------------------------------------------------------- #

    def run(
        self,
        df,
        udfs: list[Callable[[...], SparkColumn]] = None,
    ) -> DataFrame:
        import pyspark.sql.functions as F

        # Groupby arguments
        groupby = []
        if self.groupby_window:
            groupby += [
                F.window(
                    timeColumn=self.groupby_window.time_column,
                    windowDuration=self.groupby_window.window_duration,
                    slideDuration=self.groupby_window.slide_duration,
                    startTime=self.groupby_window.start_time,
                )
            ]

        for c in self.groupby_columns:
            groupby += [c]

        # Agg arguments
        aggs = []
        for expr in self.agg_expressions:
            expr.type = "_any"
            aggs += [
                expr.to_spark(
                    df=df,
                    udfs=udfs,
                    raise_exception=True,
                ).alias(expr.name)
            ]

        return df.groupby(groupby).agg(*aggs)



from typing import Union
import pyspark.sql.functions as F

# from pyspark.sql.functions import pandas_udf
from pyspark.sql.column import Column
from laktory.spark.functions._common import (
    COLUMN_OR_NAME,
    INT_OR_COLUMN,
    FLOAT_OR_COLUMN,
    STRING_OR_COLUMN,
    _col,
    _lit,
)

__all__ = [
    "poly1",
    "poly2",
    "power",
    "roundp",
]


# --------------------------------------------------------------------------- #
# Polynomials                                                                 #
# --------------------------------------------------------------------------- #


def poly1(
    x: COLUMN_OR_NAME,
    a: FLOAT_OR_COLUMN = 1.0,
    b: FLOAT_OR_COLUMN = 0.0,
) -> Column:
    """
    Polynomial function of first degree

    Parameters
    ----------
    x: pyspark.sql.functions.column.Column, column name
        Input column
    a: float or pyspark.sql.functions.column.Column
        Slope
    b: float or pyspark.sql.functions.column.Column
        y-intercept

    Returns
    -------
    output: pyspark.sql.functions.column.Column
        Output column
    """

    return _lit(a) * _col(x) + _lit(b)


def poly2(
    x: COLUMN_OR_NAME,
    a: FLOAT_OR_COLUMN = 1.0,
    b: FLOAT_OR_COLUMN = 0.0,
    c: FLOAT_OR_COLUMN = 0.0,
) -> Column:
    """
    Polynomial function of second degree

    Parameters
    ------
    x: pyspark.sql.functions.column.Column or column name
        Input column
    a: float or pyspark.sql.functions.column.Column
        x**2 coefficient
    b: float or pyspark.sql.functions.column.Column
        x**1 coefficient
    c: float or pyspark.sql.functions.column.Column
        x**0 coefficient

    Returns
    -------
    output: pyspark.sql.functions.column.Column
        Output column

    Examples
    --------
    >>> 1+1.0
    2.0
    """
    return _lit(a) * _col(x) ** 2 + _lit(b) * _col(x) + _lit(c)


# --------------------------------------------------------------------------- #
# Power                                                                       #
# --------------------------------------------------------------------------- #


def power(
    x: COLUMN_OR_NAME,
    a: FLOAT_OR_COLUMN = 1.0,
    n: FLOAT_OR_COLUMN = 0.0,
) -> Column:
    """
    Polynomial function of first degree

    Parameters
    ------
    x: pyspark.sql.functions.column.Column or column name
        Input column
    a: float or pyspark.sql.functions.column.Column
        Coefficient
    n: float or pyspark.sql.functions.column.Column
        Exponent

    Returns
    -------
    output: pyspark.sql.functions.column.Column
        Output column
    """
    return _lit(a) * _col(x) ** _lit(n)


# --------------------------------------------------------------------------- #
# Rounding                                                                    #
# --------------------------------------------------------------------------- #


def roundp(
    x: COLUMN_OR_NAME,
    p: FLOAT_OR_COLUMN = 1.0,
) -> Column:
    """
    Evenly round to the given precision

    Parameters
    ------
    x: pyspark.sql.functions.column.Column or column name
        Input column
    p: float or pyspark.sql.functions.column.Column
        Precision

    Returns
    -------
    output: pyspark.sql.functions.column.Column
        Output column
    """
    # eps0 = 1.0e-16
    # precision = float(precision)
    # if precision < eps0:
    #     raise ValueError("Precision must be greater than 1.0e-16")
    return F.round(_col(x) / _lit(p)) * _lit(p)



import os
from laktory.models import Job

root_dir = os.path.dirname(__file__)


def test_job():
    job = Job(
        name="job-stock-prices",
        clusters=[
            {
                "name": "main",
                "spark_version": "14.0.x-scala2.12",
                "node_type_id": "Standard_DS3_v2",
            },
        ],
        tasks=[
            {
                "job_cluster_key": "main",
                "notebook_task": {
                    "notebook_path": "job/ingest_stock_prices",
                },
                "task_key": "ingestion",
            },
            {
                "depends_ons": [{"task_key": "ingestion"}],
                "pipeline_task": {"pipeline_id": "TBD"},
                "task_key": "pipeline",
            },
        ],
    )


if __name__ == "__main__":
    test_job()



import yaml
import json
import jsonref
from typing import Any
from pydantic import BaseModel as _BaseModel
from pydantic import ConfigDict

from laktory.workspaceclient import WorkspaceClient


class BaseModel(_BaseModel):
    model_config = ConfigDict(extra="forbid")
    # _vars: dict[str, Any] = {}

    def __init__(self, _vars=None, **kwargs):
        super().__init__(**kwargs)
        # Setting a private field does not seem to be propagated to all child
        # classes. Using the init instead.
        if _vars is None:
            _vars = {}
        self._vars = _vars

    @property
    def workspace_client(self):
        return WorkspaceClient()

    @property
    def vars(self) -> dict[str, Any]:
        return self._vars

    @vars.setter
    def vars(self, value):
        self._vars = value

    @classmethod
    def model_validate_yaml(cls, fp):
        data = yaml.safe_load(fp)
        return cls.model_validate(data)

    @classmethod
    def model_validate_json_file(cls, fp):
        data = json.load(fp)
        return cls.model_validate(data)

    #
    # def model_pulumi_dump(self):
    #     return self.model_dump()

    @property
    def pulumi_excludes(self) -> list[str]:
        return []

    @property
    def pulumi_renames(self) -> dict[str, str]:
        return {}

    def inject_vars(self, d) -> dict:
        def search_and_replace(d, old_value, new_val):
            if isinstance(d, dict):
                for key, value in d.items():
                    d[key] = search_and_replace(value, old_value, new_val)
            elif isinstance(d, list):
                for i, item in enumerate(d):
                    d[i] = search_and_replace(item, old_value, new_val)
            elif d == old_value:
                d = new_val
            elif isinstance(d, str) and old_value in d:
                d = d.replace(old_value, new_val)
            return d

        for var_key, var_value in self.vars.items():
            d = search_and_replace(d, f"${{var.{var_key}}}", var_value)
        return d

    def model_pulumi_dump(self, *args, **kwargs):
        kwargs["exclude"] = self.pulumi_excludes
        d = super().model_dump(*args, **kwargs)
        for k, v in self.pulumi_renames.items():
            d[v] = d.pop(k)
        d = self.inject_vars(d)
        return d

    # TODO: Migrate to Databricks engine
    # @classmethod
    # def model_serialized_types(cls):
    #     def parse(schema_data):
    #         data_type = schema_data.get("type", None)
    #         all_of = schema_data.get("allOf", None)
    #         any_of = schema_data.get("anyOf", None)
    #
    #         if data_type is None and all_of is not None:
    #             return parse(all_of[0])
    #
    #         elif data_type is None and any_of is not None:
    #             return parse(any_of[0])
    #
    #         elif data_type == "object":
    #             properties = schema_data.get("properties", {})
    #             prop_mapping = {}
    #             for prop, prop_data in properties.items():
    #                 prop_mapping[prop] = parse(prop_data)
    #             return prop_mapping
    #
    #         elif data_type == "array":
    #             items = schema_data.get("items", {})
    #             return [parse(items)]
    #
    #         else:
    #             return data_type
    #
    #     json_schema = jsonref.loads(json.dumps(cls.model_json_schema()))
    #
    #     types = parse(json_schema)
    #
    #     return types

    # TODO: Migrate to Databricks engine
    # @classmethod
    # def model_sql_schema(cls, types: dict = None):
    #     if types is None:
    #         types = cls.model_serialized_types()
    #
    #     schema = (
    #         "("
    #         + ", ".join(f"{k} {py_to_sql(v, mode='schema')}" for k, v in types.items())
    #         + ")"
    #     )
    #
    #     if "<>" in schema:
    #         raise ValueError(
    #             f"Some types are undefined sql schema can't be defined\n{schema}"
    #         )
    #
    #     if "null" in schema:
    #         raise ValueError(
    #             f"Some types are undefined sql schema can't be defined\n{schema}"
    #         )
    #
    #     return schema



from typing import Literal
from laktory.models.base import BaseModel


class MetastoreGrant(BaseModel):
    principal: str
    privileges: list[
        Literal[
            "CREATE_CATALOG",
            "CREATE_EXTERNAL",
            "LOCATION",
            "CREATE_CONNECTION",
            "CREATE_RECIPIENT",
            "CREATE_SHARE",
            "CREATE_PROVIDER",
            "USE_MARKETPLACE_ASSETS",
            "USE_PROVIDER",
            "USE_SHARE",
            "USE_RECIPIENT",
            "SET_SHARE_PERMISSION",
            "MANAGE_ALLOWLIST",
        ]
    ]



from pydantic import field_validator
from pydantic import model_validator
from typing import Any
from typing import Callable
from typing import Union

from laktory._logger import get_logger
from laktory.contants import SUPPORTED_TYPES
from laktory.models.base import BaseModel
from laktory.spark import Column as SparkColumn

logger = get_logger(__name__)


class SparkFuncArg(BaseModel):
    value: Any
    is_column: bool = None
    to_lit: bool = None

    @model_validator(mode="after")
    def is_column_default(self) -> Any:
        if self.to_lit is None and self.is_column is None:
            if isinstance(self.value, str):
                self.is_column = True

        if self.is_column is None:
            self.is_column = False

        if self.to_lit is None:
            self.to_lit = False

        if self.is_column and self.to_lit:
            raise ValueError("Only one of `to_column` or `to_lit` can be True")

        return self

    def to_spark(self):
        import pyspark.sql.functions as F

        v = self.value
        if self.to_lit:
            v = F.lit(v)
        else:
            v = F.expr(v)
        return v


class Column(BaseModel):
    catalog_name: Union[str, None] = None
    comment: Union[str, None] = None
    name: str
    pii: Union[bool, None] = None
    schema_name: Union[str, None] = None
    spark_func_args: list[Union[str, SparkFuncArg]] = []
    spark_func_kwargs: dict[str, Union[Any, SparkFuncArg]] = {}
    spark_func_name: Union[str, None] = None
    sql_expression: Union[str, None] = None
    table_name: Union[str, None] = None
    type: str = "string"
    unit: Union[str, None] = None

    @field_validator("spark_func_args")
    def parse_args(cls, args: list[Union[str, SparkFuncArg]]) -> list[SparkFuncArg]:
        _args = []
        for a in args:
            if isinstance(a, SparkFuncArg):
                pass
            elif isinstance(a, dict):
                a = SparkFuncArg(**a)
            else:
                a = SparkFuncArg(value=a)
            _args += [a]
        return _args

    @field_validator("spark_func_kwargs")
    def parse_kwargs(
        cls, kwargs: dict[str, Union[str, SparkFuncArg]]
    ) -> dict[str, SparkFuncArg]:
        _kwargs = {}
        for k, a in kwargs.items():
            if isinstance(a, SparkFuncArg):
                pass
            elif isinstance(a, dict):
                a = SparkFuncArg(**a)
            else:
                a = SparkFuncArg(value=a)
            _kwargs[k] = a
        return _kwargs

    @field_validator("type")
    def default_load_path(cls, v: str) -> str:
        if "<" in v:
            return v
        else:
            if v not in SUPPORTED_TYPES:
                raise ValueError(
                    f"Type {v} is not supported. Select one of {SUPPORTED_TYPES}"
                )
        return v

    # ----------------------------------------------------------------------- #
    # Computed fields                                                         #
    # ----------------------------------------------------------------------- #

    @property
    def parent_full_name(self) -> str:
        _id = ""
        if self.catalog_name:
            _id += self.catalog_name

        if self.schema_name:
            if _id == "":
                _id = self.schema_name
            else:
                _id += f".{self.schema_name}"

        if self.table_name:
            if _id == "":
                _id = self.table_name
            else:
                _id += f".{self.table_name}"

        return _id

    @property
    def full_name(self) -> str:
        _id = self.name
        if self.parent_full_name is not None:
            _id = f"{self.parent_full_name}.{_id}"
        return _id

    @property
    def database_name(self) -> str:
        return self.schema_name

    # ----------------------------------------------------------------------- #
    # Class Methods                                                           #
    # ----------------------------------------------------------------------- #

    def to_spark(
        self,
        df,
        udfs: list[Callable[[...], SparkColumn]] = None,
        raise_exception: bool = True,
    ) -> SparkColumn:
        import pyspark.sql.functions as F
        from laktory.spark import functions as LF
        from laktory.spark.dataframe import has_column

        if udfs is None:
            udfs = []
        udfs = {f.__name__: f for f in udfs}

        # From SQL expression
        if self.sql_expression:
            logger.info(f"   {self.name}[{self.type}] as `{self.sql_expression}`)")
            col = F.expr(self.sql_expression).alias(self.name)
            if self.type not in ["_any"]:
                col = col.cast(self.type)
            return col

        # From Spark Function
        func_name = self.spark_func_name
        if func_name is None:
            func_name = "coalesce"

        # Get from UDFs
        f = udfs.get(func_name, None)
        if f is None:
            # Get from built-in spark functions
            f = getattr(F, func_name, None)

            if f is None:
                # Get from laktory functions
                f = getattr(LF, func_name, None)

        if f is None:
            raise ValueError(f"Function {func_name} is not available")

        _args = self.spark_func_args
        _kwargs = self.spark_func_kwargs

        logger.info(f"   {self.name}[{self.type}] as {func_name}({_args}, {_kwargs})")

        # Build args
        args = []
        expected_cols = 0
        found_cols = 0
        for i, _arg in enumerate(_args):
            if df is not None:
                if _arg.is_column:
                    expected_cols += 1
                    if not has_column(df, _arg.value):
                        # When columns are not found, they are simply skipped and a
                        # warning is issued. Some functions, like `coalesce` might list
                        # multiple arguments, but don't expect all of them to be
                        # available
                        logger.warning(f"Column '{_arg.value}' not available")
                        continue
                    else:
                        found_cols += 1

            # TODO: Review if required
            # if _arg.value.startswith("data.") or func_name == "coalesce":
            #     pass
            # input_type = dict(df.dtypes)[input_col_name]
            # if input_type in ["double"]:
            #     # Some bronze NaN data will be converted to 0 if cast to int
            #     input_col = F.when(F.isnan(input_col_name), None).otherwise(F.col(input_col_name))
            # if self.type not in ["_any"] and func_name not in [
            #     "to_safe_timestamp"
            # ]:
            #     input_col = F.col(input_col_name).cast(self.type)

            args += [_arg.to_spark()]

        if expected_cols > 0 and found_cols == 0:
            if raise_exception:
                raise ValueError(
                    f"None of the inputs columns ({_args}) for {self.name} have been found"
                )
            else:
                logger.info("Input columns not available. Skipping")
                return None

        # Build kwargs
        kwargs = {}
        for k, _arg in _kwargs.items():
            kwargs[k] = _arg.to_spark()

        # Function call
        col = f(*args, **kwargs)

        # Type Casting
        if self.type not in ["_any"]:
            col = col.cast(self.type)

        return col



from typing import Literal
from typing import Union
from typing import Any
from pydantic import BaseModel as _BaseModel

from laktory._settings import settings

# from laktory.resourcesengines.pulumi.base import PulumiResourcesEngine
# from laktory.resourcesengines.databricks.base import DatabricksResourcesEngine

ENGINES = ["pulumi", "databricks-api", "terraform"]


class Resources(_BaseModel):
    # _resources: Union[PulumiResourcesEngine, DatabricksResourcesEngine] = None
    _resources: Any = None

    @property
    def resources(self):
        if self._resources is None:
            raise ValueError(
                f"Model ({self}) has not been deployed. Call model.deploy() first"
            )
        return self._resources

    def deploy(self, *args, engine: Literal[tuple(ENGINES)] = None, **kwargs):
        if not engine:
            engine = settings.resources_engine
        engine = engine.lower()

        if engine == "pulumi":
            self._resources = self.deploy_with_pulumi(*args, **kwargs)
        elif engine == "databricks-api":
            raise NotImplementedError(
                "Databricks API deployments are not yet supported"
            )
            self._resources = self.deploy_with_databricks(*args, **kwargs)
        elif engine == "terraform":
            raise NotImplementedError("Terraform deployments are not yet supported")
            self._resources = self.deploy_with_terraform(*args, **kwargs)
        else:
            raise ValueError(
                f"Engine {engine} is not supported. Available engines: {ENGINES}"
            )

        return self._resources

    def deploy_with_pulumi(self, *args, **kwargs):
        raise NotImplementedError()

    def deploy_with_databricks(self, *args, **kwargs):
        raise NotImplementedError()

    def deploy_with_terraform(self, *args, **kwargs):
        raise NotImplementedError()



#
#
# # --------------------------------------------------------------------------- #
# # to_safe_timestamp                                                           #
# # --------------------------------------------------------------------------- #
#
# @pandas_udf(sqlt.TimestampType(), udfuncs)
# def to_safe_timestamp(x: pd.Series, utc: bool=True) -> pd.Series:
#     """
#     Convert to timestamp when input data is a mix of:
#         - Timestamp string
#         - Unix timestamps in string (including exponential notation)
#         - Unix timestamps as floats
#
#     Parameters
#     ------
#     x: pd.Series
#         Input data
#
#     Returns
#     -------
#     output: pd.Series
#         Output series
#     """
#     x = x.copy()
#
#     # Find NaN
#     _x = x.str.lower().str
#     where_na = _x.contains("nan") | _x.contains("n/a") | _x.contains("null")
#
#     # Convert strings to datetime
#     where = x.str.contains("-") | x.str.contains(":") & ~where_na
#     _s = x.loc[where]
#     _s = pd.to_datetime(_s, utc=utc)
#
#     # Convert datetimes to unix timestamps
#     x.loc[where] = (_s - datetime(1970, 1, 1, tzinfo=pytz.utc)).dt.total_seconds().astype(pd.Float64Dtype())
#
#     # Set NaN
#     x.loc[where_na] = None
#
#     # Convert everything to datetimes
#     x = pd.to_datetime(x, unit="s")
#
#     return x
#
#


# # --------------------------------------------------------------------------- #
# # to_safe_timestamp                                                           #
# # --------------------------------------------------------------------------- #
#
# @pandas_udf(sqlt.TimestampType(), udfuncs)
# def to_timestamp(x: pd.Series) -> pd.Series:
#     """
#     Convert to timestamp when input data is unix timestamps as floats
#
#     Parameters
#     ------
#     x: pd.Series
#         Input data
#
#     Returns
#     -------
#     output: pd.Series
#         Output series
#     """
#     # Convert everything to datetimes
#     x = pd.to_datetime(x, unit="s")
#     return x
#
#
# # --------------------------------------------------------------------------- #
# # unix_timestamp                                                              #
# # --------------------------------------------------------------------------- #
#
# @pandas_udf(sqlt.DoubleType(), udfuncs)
# def unix_timestamp(
#     x: pd.Series,
# ) -> pd.Series:
#     """
#     Return current unix timestamp in seconds
#
#     Parameters
#     ----------
#     x: pd.Series, default = None
#         Series used to set index
#
#     Returns
#     -------
#     output: float, pd.Series
#         Current timestamp
#     """
#     return pd.Series(_unix_timestamp(), index=x.index)
#
#



from laktory.resourcesengines import PulumiResourcesEngine


class PulumiResources(PulumiResourcesEngine):
    @property
    def provider(self) -> str:
        return "databricks"


def test_pulumi_engine():
    # TODO: Setup pulumi and deploy resource
    resources = PulumiResources("t", "name")
    assert resources.t == "laktory:databricks:Resources"


if __name__ == "__main__":
    test_pulumi_engine()



from laktory.models.compute.workspacefile import WorkspaceFile
from laktory.models.compute.cluster import Cluster
from laktory.models.compute.warehouse import Warehouse
from laktory.models.compute.notebook import Notebook
from laktory.models.compute.pipeline import Pipeline
from laktory.models.compute.tablejoin import TableJoin
from laktory.models.compute.tableaggregation import TableAggregation
from laktory.models.compute.tablebuilder import TableBuilder
from laktory.models.compute.tablewindowfilter import TableWindowFilter
from laktory.models.compute.job import Job



from typing import Union
import pyspark.sql.functions as F

# from pyspark.sql.functions import pandas_udf
from pyspark.sql.column import Column
from laktory.spark.functions._common import (
    COLUMN_OR_NAME,
    INT_OR_COLUMN,
    FLOAT_OR_COLUMN,
    STRING_OR_COLUMN,
    _col,
    _lit,
)

__all__ = [
    "string_split",
    "uuid",
]


# --------------------------------------------------------------------------- #
# string_split                                                                #
# --------------------------------------------------------------------------- #


def string_split(
    x: COLUMN_OR_NAME,
    pattern: str,
    key: int,
) -> Column:
    """
    Get substring using separator `pat`.

    Parameters
    ------
    x: pyspark.sql.functions.column.Column or column name
        Input text series to split
    pattern: str or pyspark.sql.functions.column.Column
        String or regular expression to split on. If not specified, split on whitespace.
    key: int or pyspark.sql.functions.column.Column
        Split index to return

    Returns
    -------
    output: pd.Series
        Output series
    """
    return F.split(_col(x), pattern=pattern).getItem(key)


# --------------------------------------------------------------------------- #
# uuid                                                                        #
# --------------------------------------------------------------------------- #


def uuid() -> Column:
    """
    Create a unique id for each row.

    Parameters
    ------

    Returns
    -------
    output: pyspark.sql.functions.column.Column
        Output column
    """
    return F.expr("uuid()")



from datetime import datetime

from laktory.models import Table
from laktory.models import Schema
from laktory.models import Catalog
from laktory.models import Column


def test_model():
    db = Schema(
        name="flights",
        catalog_name="laktory_testing",
        tables=[
            Table(
                name="f1549",
                columns=[
                    {
                        "name": "airspeed",
                        "type": "double",
                    },
                    {
                        "name": "altitude",
                        "type": "double",
                    },
                ],
            ),
            Table(
                name="f0002",
                columns=[
                    {
                        "name": "airspeed",
                        "type": "double",
                    },
                    {
                        "name": "altitude",
                        "type": "double",
                    },
                ],
            ),
        ],
    )

    assert db.tables[0].columns[0].name == "airspeed"
    assert type(db.tables[0].columns[0]) == Column
    assert db.name == "flights"
    assert db.full_name == "laktory_testing.flights"


def atest_create():
    # Timestamp is included in catalog name to prevent conflicts when running
    # multiple tests in parallel
    catalog_name = "laktory_testing_" + str(datetime.now().timestamp()).replace(".", "")

    cat = Catalog(name=catalog_name)
    cat.create(if_not_exists=True)
    db = Schema(name="default", catalog_name=catalog_name)
    db.create()
    assert db.exists()
    db.delete()
    assert not db.exists()
    cat.delete(force=True)


if __name__ == "__main__":
    test_model()
    # atest_create()



from typing import Union
from laktory.models.base import BaseModel
from laktory.models.resources import Resources


class User(BaseModel, Resources):
    user_name: str
    display_name: str = None
    workspace_access: bool = None
    groups: list[str] = []
    roles: list[str] = []
    id: Union[str, None] = None

    # Deployment options
    disable_as_user_deletion: bool = False

    # ----------------------------------------------------------------------- #
    # Resources Engine Methods                                                #
    # ----------------------------------------------------------------------- #
    @property
    def pulumi_excludes(self) -> list[str]:
        return ["groups", "roles", "id"]

    def deploy_with_pulumi(self, name=None, group_ids=None, opts=None):
        from laktory.resourcesengines.pulumi.user import PulumiUser

        return PulumiUser(name=name, user=self, group_ids=group_ids, opts=opts)



from typing import Literal
from laktory.models.base import BaseModel

PRIVILEGES = [
    "ALL PRIVILEGES",
    "CREATE_EXTERNAL_LOCATION",
    "CREATE_EXTERNAL_TABLE",
    "READ_FILES",
    "WRITE_FILES",
]


class StorageCredentialGrant(BaseModel):
    principal: str
    privileges: list[Literal[tuple(PRIVILEGES)]]



import pulumi
import pulumi_databricks as databricks
from laktory.resourcesengines.pulumi.base import PulumiResourcesEngine
from laktory.models.secrets.secretscope import SecretScope

from laktory._logger import get_logger

logger = get_logger(__name__)


class PulumiSecretScope(PulumiResourcesEngine):
    @property
    def provider(self):
        return "databricks"

    def __init__(
        self,
        name=None,
        secret_scope: SecretScope = None,
        opts=None,
    ):
        if name is None:
            name = f"secret-scope-{secret_scope.name}"
        super().__init__(self.t, name, {}, opts)

        opts = pulumi.ResourceOptions(
            parent=self,
            # delete_before_replace=True,
        )

        self.secret_scope = databricks.SecretScope(
            f"secret-scope-{secret_scope.name}",
            opts=opts,
            **secret_scope.model_pulumi_dump(),
        )

        for s in secret_scope.secrets:
            databricks.Secret(
                f"secret-{secret_scope.name}-{s.key}",
                key=s.key,
                string_value=s.value,
                scope=self.secret_scope.id,
                opts=opts,
            )

        for p in secret_scope.permissions:
            databricks.SecretAcl(
                f"secret-scope-acl-{secret_scope.name}-{p.principal}",
                permission=p.permission,
                principal=p.principal,
                scope=self.secret_scope.name,
                opts=opts,
            )
